#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

from neutron.api import extensions
from neutron.api.v2 import resource as neutron_resource
from neutron.db import l3_db
from neutron.db import models_v2
from neutron.db import securitygroups_db
from neutron.objects import securitygroup as sg_obj
from neutron.plugins.ml2 import db as ml2_db
from neutron.quota import resource as quota_resource
from neutron_lib.api import attributes
from neutron_lib.api import validators
from neutron_lib.db import model_query
from neutron_lib import exceptions
from neutron_lib.plugins import directory
from oslo_log import log
from oslo_utils import excutils

from gbpservice.common import utils as gbp_utils
from gbpservice.neutron.db import api as db_api


LOG = log.getLogger(__name__)


if not hasattr(quota_resource, 'GBP_PATCHED'):
    orig_count_resource = quota_resource._count_resource

    def new_count_resource(*kwargs):
        request = gbp_utils.get_obj_from_stack(neutron_resource.Request)
        orig_plugins = directory._get_plugin_directory()._plugins
        if request and request.environ['PATH_INFO'] == (
                '/servicechain/service_profiles.json'):
            new_plugins = copy.copy(directory._get_plugin_directory()._plugins)
            # The service_profile resource is supported by the FLAVORS
            # plugin as well as the SERVICECHAIN plugin. At this point
            # we know that we are dealing with the service_profile from
            # SERVICECHAIN, and since the original implementation of the
            # count_resource will think of service_profile from FLAVORS
            # (in the sorted order of plugins, FLAVORS preceedes SERVICECHAIN)
            # we temporarily remove the FLAVORS plugin reference from the
            # plugins directory.
            new_plugins.pop('FLAVORS')
            directory._get_plugin_directory()._plugins = new_plugins

        count_resource = orig_count_resource(*kwargs)
        directory._get_plugin_directory()._plugins = orig_plugins
        return count_resource

    quota_resource._count_resource = new_count_resource
    quota_resource.GBP_PATCHED = True


# REVISIT(ivar): Monkey patch to allow explicit router_id to be set in Neutron
# for Floating Ip creation (for internal calls only). Once we split the server,
# this could be part of a GBP Neutron L3 driver.
def _get_assoc_data(self, context, fip, floatingip_db):
    (internal_port, internal_subnet_id,
     internal_ip_address) = self._internal_fip_assoc_data(
         context, fip, floatingip_db['tenant_id'])
    if fip.get('router_id'):
        router_id = fip['router_id']
        del fip['router_id']
    else:
        router_id = self._get_router_for_floatingip(
            context, internal_port, internal_subnet_id,
            floatingip_db['floating_network_id'])

    return fip['port_id'], internal_ip_address, router_id


l3_db.L3_NAT_dbonly_mixin._get_assoc_data = _get_assoc_data


# REVISIT(ivar): Neutron adds a tenant filter on SG lookup for a given port,
# this breaks our service chain plumbing model so for now we should monkey
# patch the specific method. A follow up with the Neutron team is needed to
# figure out the reason for this and how to proceed for future releases.
def _get_security_groups_on_port(self, context, port):
    """Check that all security groups on port belong to tenant.

    :returns: all security groups IDs on port belonging to tenant.
    """
    p = port['port']
    if not validators.is_attr_set(
            p.get(securitygroups_db.ext_sg.SECURITYGROUPS)):
        return
    if p.get('device_owner') and p['device_owner'].startswith('network:'):
        return

    port_sg = p.get(securitygroups_db.ext_sg.SECURITYGROUPS, [])
    sg_objs = sg_obj.SecurityGroup.get_objects(context, id=port_sg)
    valid_groups = set(g['id'] for g in sg_objs)

    requested_groups = set(port_sg)
    port_sg_missing = requested_groups - valid_groups
    if port_sg_missing:
        raise securitygroups_db.ext_sg.SecurityGroupNotFound(
            id=', '.join(port_sg_missing))

    return sg_objs


securitygroups_db.SecurityGroupDbMixin._get_security_groups_on_port = (
    _get_security_groups_on_port)


def get_port_from_device_mac(context, device_mac):
    LOG.debug("get_port_from_device_mac() called for mac %s", device_mac)
    qry = context.session.query(models_v2.Port).filter_by(
        mac_address=device_mac).order_by(models_v2.Port.device_owner.desc())
    return qry.first()


ml2_db.get_port_from_device_mac = get_port_from_device_mac


def extend_resources(self, version, attr_map):
    """Extend resources with additional resources or attributes.

    :param attr_map: the existing mapping from resource name to
    attrs definition.

    After this function, we will extend the attr_map if an extension
    wants to extend this map.
    """
    # Match indentation of original version, making comparison easier.
    if True:
        processed_exts = {}
        exts_to_process = self.extensions.copy()
        check_optionals = True
        # Iterate until there are unprocessed extensions or if no progress
        # is made in a whole iteration
        while exts_to_process:
            processed_ext_count = len(processed_exts)
            for ext_name, ext in list(exts_to_process.items()):
                # Process extension only if all required extensions
                # have been processed already
                required_exts_set = set(ext.get_required_extensions())
                if required_exts_set - set(processed_exts):
                    continue
                optional_exts_set = set(ext.get_optional_extensions())
                if check_optionals and optional_exts_set - set(processed_exts):
                    continue
                extended_attrs = ext.get_extended_resources(version)
                for res, resource_attrs in extended_attrs.items():
                    res_to_update = attr_map.setdefault(res, {})
                    if self._is_sub_resource(res_to_update):
                        # kentwu: service_profiles defined in servicechain
                        # plugin has a name conflict with service_profiles
                        # sub-resource defined in flavor plugin. The attr_map
                        # can only have one service_profiles so here we make
                        # this very same service_profiles to have the
                        # attributes from both plugins. This behavior is now
                        # consistent with Pike.
                        if (ext_name == 'servicechain' and
                                res == 'service_profiles'):
                            res_to_update.update(resource_attrs)
                        # in the case of an existing sub-resource, we need to
                        # update the parameters content rather than overwrite
                        # it, and also keep the description of the parent
                        # resource unmodified
                        else:
                            res_to_update['parameters'].update(
                                resource_attrs['parameters'])
                    else:
                        res_to_update.update(resource_attrs)
                processed_exts[ext_name] = ext
                del exts_to_process[ext_name]
            if len(processed_exts) == processed_ext_count:
                # if we hit here, it means there are unsatisfied
                # dependencies. try again without optionals since optionals
                # are only necessary to set order if they are present.
                if check_optionals:
                    check_optionals = False
                    continue
                # Exit loop as no progress was made
                break
        if exts_to_process:
            unloadable_extensions = set(exts_to_process.keys())
            LOG.error("Unable to process extensions (%s) because "
                      "the configured plugins do not satisfy "
                      "their requirements. Some features will not "
                      "work as expected.",
                      ', '.join(unloadable_extensions))
            self._check_faulty_extensions(unloadable_extensions)
        # Extending extensions' attributes map.
        for ext in processed_exts.values():
            ext.update_attributes_map(attr_map)


extensions.ExtensionManager.extend_resources = extend_resources


def fill_post_defaults(
        self, res_dict,
        exc_cls=lambda m: exceptions.InvalidInput(error_message=m),
        check_allow_post=True):
    """Fill in default values for attributes in a POST request.

    When a POST request is made, the attributes with default values do not
    need to be specified by the user. This function fills in the values of
    any unspecified attributes if they have a default value.

    If an attribute is not specified and it does not have a default value,
    an exception is raised.

    If an attribute is specified and it is not allowed in POST requests, an
    exception is raised. The caller can override this behavior by setting
    check_allow_post=False (used by some internal admin operations).

    :param res_dict: The resource attributes from the request.
    :param exc_cls: Exception to be raised on error that must take
        a single error message as it's only constructor arg.
    :param check_allow_post: Raises an exception if a non-POST-able
        attribute is specified.
    :raises: exc_cls If check_allow_post is True and this instance of
        ResourceAttributes doesn't support POST.
    """
    for attr, attr_vals in self.attributes.items():
        # kentwu: Patch needed for our GBP service_profiles attribute. Since
        # parent and parameters are both sub-resource's attributes picked up
        # from flavor plugin so we can just ignore those. These 2 attributes
        # don't have allow_post defined so it will just fail without this
        # patch.
        if attr == 'parent' or attr == 'parameters':
            if 'allow_post' not in attr_vals:
                continue
        if attr_vals['allow_post']:
            if 'default' not in attr_vals and attr not in res_dict:
                msg = _("Failed to parse request. Required "
                        "attribute '%s' not specified") % attr
                raise exc_cls(msg)
            res_dict[attr] = res_dict.get(attr,
                                          attr_vals.get('default'))
        elif check_allow_post:
            if attr in res_dict:
                msg = _("Attribute '%s' not allowed in POST") % attr
                raise exc_cls(msg)


attributes.AttributeInfo.fill_post_defaults = fill_post_defaults


# TODO(ivar): while this block would be better place in the patch_neutron
# module, it seems like being part of an "extension" package is the only
# way to make it work at the moment. Tests have shown that Neutorn reloads
# the extensions at every call (at least in the UTs) and this causes the
# AIM_FLC_L7_PARAMS to be reset over and over. By patching at this point,
# we make sure we always have the proper value for that variable.
try:
    import six
    import sys

    from networking_sfc.db import flowclassifier_db
    from networking_sfc.db import sfc_db
    from networking_sfc.extensions import flowclassifier as fc_ext
    from networking_sfc.extensions import sfc as sfc_ext  # noqa
    from networking_sfc.services.flowclassifier.common import (
        exceptions as fc_exc)
    from networking_sfc.services.flowclassifier import driver_manager as fc_mgr
    from networking_sfc.services.flowclassifier import plugin as fc_plugin
    from networking_sfc.services.sfc.common import exceptions as sfc_exc
    from networking_sfc.services.sfc import driver_manager as sfc_mgr
    from networking_sfc.services.sfc import plugin as sfc_plugin
    from neutron_lib.services.trunk import constants
    from oslo_utils import uuidutils

    from gbpservice.neutron.services.sfc.aim import constants as sfc_cts

    if 'flowclassifier' in sys.modules:
        sys.modules['flowclassifier'].SUPPORTED_L7_PARAMETERS.update(
            sfc_cts.AIM_FLC_L7_PARAMS)
    if 'networking_sfc.extensions.flowclassifier' in sys.modules:
        sys.modules[
            ('networking_sfc.extensions.'
             'flowclassifier')].SUPPORTED_L7_PARAMETERS.update(
            sfc_cts.AIM_FLC_L7_PARAMS)
    if 'sfc' in sys.modules:
        sys.modules['sfc'].RESOURCE_ATTRIBUTE_MAP['port_pair_groups'][
            'port_pair_group_parameters']['validate']['type:dict'].update(
                sfc_cts.AIM_PPG_PARAMS)
    if 'networking_sfc.extensions.sfc' in sys.modules:
        sys.modules['networking_sfc.extensions.sfc'].RESOURCE_ATTRIBUTE_MAP[
            'port_pair_groups']['port_pair_group_parameters']['validate'][
            'type:dict'].update(sfc_cts.AIM_PPG_PARAMS)
    # REVISIT(ivar): The following diff will fix flow classifier creation
    # method when using L7 parameters.
    # -            key: L7Parameter(key, val)
    # +            key: L7Parameter(keyword=key, value=val)

    def create_flow_classifier(self, context, flow_classifier):
        fc = flow_classifier['flow_classifier']
        tenant_id = fc['tenant_id']
        l7_parameters = {
            key: flowclassifier_db.L7Parameter(keyword=key, value=val)
            for key, val in six.iteritems(fc['l7_parameters'])}
        ethertype = fc['ethertype']
        protocol = fc['protocol']
        source_port_range_min = fc['source_port_range_min']
        source_port_range_max = fc['source_port_range_max']
        self._check_port_range_valid(source_port_range_min,
                                     source_port_range_max,
                                     protocol)
        destination_port_range_min = fc['destination_port_range_min']
        destination_port_range_max = fc['destination_port_range_max']
        self._check_port_range_valid(destination_port_range_min,
                                     destination_port_range_max,
                                     protocol)
        source_ip_prefix = fc['source_ip_prefix']
        self._check_ip_prefix_valid(source_ip_prefix, ethertype)
        destination_ip_prefix = fc['destination_ip_prefix']
        self._check_ip_prefix_valid(destination_ip_prefix, ethertype)
        logical_source_port = fc['logical_source_port']
        logical_destination_port = fc['logical_destination_port']
        with db_api.CONTEXT_WRITER.using(context):
            if logical_source_port is not None:
                self._get_port(context, logical_source_port)
            if logical_destination_port is not None:
                self._get_port(context, logical_destination_port)
            query = model_query.query_with_hooks(
                context, flowclassifier_db.FlowClassifier)
            for flow_classifier_db in query.all():
                if self.flowclassifier_conflict(
                    fc,
                    flow_classifier_db
                ):
                    raise fc_ext.FlowClassifierInConflict(
                        id=flow_classifier_db['id'])
            flow_classifier_db = flowclassifier_db.FlowClassifier(
                id=uuidutils.generate_uuid(),
                tenant_id=tenant_id,
                name=fc['name'],
                description=fc['description'],
                ethertype=ethertype,
                protocol=protocol,
                source_port_range_min=source_port_range_min,
                source_port_range_max=source_port_range_max,
                destination_port_range_min=destination_port_range_min,
                destination_port_range_max=destination_port_range_max,
                source_ip_prefix=source_ip_prefix,
                destination_ip_prefix=destination_ip_prefix,
                logical_source_port=logical_source_port,
                logical_destination_port=logical_destination_port,
                l7_parameters=l7_parameters
            )
            context.session.add(flow_classifier_db)
            return self._make_flow_classifier_dict(flow_classifier_db)
    flowclassifier_db.FlowClassifierDbPlugin.create_flow_classifier = (
        create_flow_classifier)

    # Flowclassifier validation should also take into account l7_parameters.

    old_validation = (
        flowclassifier_db.FlowClassifierDbPlugin.flowclassifier_basic_conflict)

    def flowclassifier_basic_conflict(cls, first_flowclassifier,
                                      second_flowclassifier):
        def _l7_params_conflict(fc1, fc2):
            if (validators.is_attr_set(fc1['l7_parameters']) and
                    validators.is_attr_set(fc2['l7_parameters'])):
                if fc1['l7_parameters'] == fc2['l7_parameters']:
                    return True
            return all(not validators.is_attr_set(fc['l7_parameters'])
                       for fc in [fc1, fc2])
        return cls._old_flowclassifier_basic_conflict(
            first_flowclassifier, second_flowclassifier) and (
            _l7_params_conflict(first_flowclassifier, second_flowclassifier))

    if getattr(flowclassifier_db.FlowClassifierDbPlugin,
               '_old_flowclassifier_basic_conflict', None) is None:
        flowclassifier_db.FlowClassifierDbPlugin.\
            _old_flowclassifier_basic_conflict = old_validation
        flowclassifier_db.FlowClassifierDbPlugin.\
            flowclassifier_basic_conflict = classmethod(
                flowclassifier_basic_conflict)

    # NOTE(ivar): Trunk subports don't have a device ID, we need this
    # validation to pass
    # NOTE(ivar): It would be ideal to re-use the original function and call
    # it instead of copying it here. However, it looks like this module gets
    # reloaded a number of times and this would cause the original definition
    # to be overridden with the new one, thus causing an endless recursion.
    def _validate_port_pair_ingress_egress(self, ingress, egress):
        if any(port.get('device_owner') == constants.TRUNK_SUBPORT_OWNER
               for port in [ingress, egress]):
            return
        if 'device_id' not in ingress or not ingress['device_id']:
            raise sfc_db.ext_sfc.PortPairIngressNoHost(
                ingress=ingress['id']
            )
        if 'device_id' not in egress or not egress['device_id']:
            raise sfc_db.ext_sfc.PortPairEgressNoHost(
                egress=egress['id']
            )
        if ingress['device_id'] != egress['device_id']:
            raise sfc_db.ext_sfc.PortPairIngressEgressDifferentHost(
                ingress=ingress['id'],
                egress=egress['id'])
    sfc_db.SfcDbPlugin._validate_port_pair_ingress_egress = (
        _validate_port_pair_ingress_egress)

    if not getattr(sfc_plugin.SfcPlugin, '_patch_db_retry', False):
        sfc_plugin.SfcPlugin.create_port_pair = (
            db_api.retry_if_session_inactive()(
                sfc_plugin.SfcPlugin.create_port_pair))
        sfc_plugin.SfcPlugin.create_port_pair_group = (
            db_api.retry_if_session_inactive()(
                sfc_plugin.SfcPlugin.create_port_pair_group))
        sfc_plugin.SfcPlugin.create_port_chain = (
            db_api.retry_if_session_inactive()(
                sfc_plugin.SfcPlugin.create_port_chain))

        sfc_plugin.SfcPlugin.update_port_pair = (
            db_api.retry_if_session_inactive()(
                sfc_plugin.SfcPlugin.update_port_pair))
        sfc_plugin.SfcPlugin.update_port_pair_group = (
            db_api.retry_if_session_inactive()(
                sfc_plugin.SfcPlugin.update_port_pair_group))
        sfc_plugin.SfcPlugin.update_port_chain = (
            db_api.retry_if_session_inactive()(
                sfc_plugin.SfcPlugin.update_port_chain))

        sfc_plugin.SfcPlugin.delete_port_pair = (
            db_api.retry_if_session_inactive()(
                sfc_plugin.SfcPlugin.delete_port_pair))
        sfc_plugin.SfcPlugin.delete_port_pair_group = (
            db_api.retry_if_session_inactive()(
                sfc_plugin.SfcPlugin.delete_port_pair_group))
        sfc_plugin.SfcPlugin.delete_port_chain = (
            db_api.retry_if_session_inactive()(
                sfc_plugin.SfcPlugin.delete_port_chain))
        sfc_plugin.SfcPlugin._patch_db_retry = True

    if not getattr(fc_plugin.FlowClassifierPlugin, '_patch_db_retry', False):
        fc_plugin.FlowClassifierPlugin.create_flow_classifier = (
            db_api.retry_if_session_inactive()(
                fc_plugin.FlowClassifierPlugin.create_flow_classifier))
        fc_plugin.FlowClassifierPlugin.update_flow_classifier = (
            db_api.retry_if_session_inactive()(
                fc_plugin.FlowClassifierPlugin.update_flow_classifier))
        fc_plugin.FlowClassifierPlugin.delete_flow_classifier = (
            db_api.retry_if_session_inactive()(
                fc_plugin.FlowClassifierPlugin.delete_flow_classifier))
        fc_plugin.FlowClassifierPlugin._patch_db_retry = True

    # Overrides SFC implementation to avoid eating retriable exceptions
    def get_call_drivers(exception, name):
        def _call_drivers(self, method_name, context, raise_orig_exc=False):
            for driver in self.ordered_drivers:
                try:
                    getattr(driver.obj, method_name)(context)
                except Exception as e:
                    # This is an internal failure.
                    if db_api.is_retriable(e):
                        with excutils.save_and_reraise_exception():
                            LOG.debug(
                                "DB exception raised by extension driver "
                                "'%(name)s' in %(method)s",
                                {'name': driver.name, 'method': method_name},
                                exc_info=e)
                    LOG.exception(e)
                    LOG.error("%(plugin)s driver '%(name)s' "
                              "failed in %(method)s",
                              {'name': driver.name, 'method': method_name,
                               'plugin': name})
                    if raise_orig_exc:
                        raise
                    else:
                        raise exception(method=method_name)
        return _call_drivers

    sfc_mgr.SfcDriverManager._call_drivers = get_call_drivers(
        sfc_exc.SfcDriverError, 'SFC')
    fc_mgr.FlowClassifierDriverManager._call_drivers = get_call_drivers(
        fc_exc.FlowClassifierDriverError, 'FlowClassifier')

except ImportError as e:
    LOG.warning("Import error while patching networking-sfc: %s",
                str(e))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from gbpservice.nfp.lib import nfp_context_manager as nfp_ctx_mgr
from gbpservice.nfp.orchestrator.coal.networking import (
    nfp_neutron_network_driver as neutron_nd
)
from gbpservice.nfp.orchestrator.openstack import openstack_driver


class NFPGBPNetworkDriver(neutron_nd.NFPNeutronNetworkDriver):

    def __init__(self, config):
        self.config = config
        super(NFPGBPNetworkDriver, self).__init__(config)
        self.network_handler = openstack_driver.GBPClient(config)

    def setup_traffic_steering(self):
        pass

    def create_port(self, token, tenant_id, net_id, name=None):
        with nfp_ctx_mgr.GBPContextManager as gcm:
            port = gcm.retry(self.network_handler.create_policy_target,
                             token, tenant_id, net_id, name)
        return port

    def delete_port(self, token, port_id):
        with nfp_ctx_mgr.GBPContextManager as gcm:
            gcm.retry(
                self.network_handler.delete_policy_target,
                token, port_id)

    def get_port_id(self, token, port_id):
        with nfp_ctx_mgr.GBPContextManager as gcm:
            pt = gcm.retry(
                self.network_handler.get_policy_target, token, port_id)
        return pt['port_id']

    def update_port(self, token, port_id, port):
        with nfp_ctx_mgr.GBPContextManager as gcm:
            pt = gcm.retry(
                self.network_handler.update_policy_target,
                token, port_id, port)
        return pt['port_id']

    def get_neutron_port_details(self, token, port_id):
        # self.network_handler = openstack_driver.NeutronClient(self.config)
        port_details = (
            super(NFPGBPNetworkDriver, self).get_port_and_subnet_details(
                token, port_id))
        # self.network_handler = openstack_driver.GBPClient(self.config)
        return port_details

    def get_port_details(self, token, port_id):
        _port_id = self.get_port_id(token, port_id)
        # self.network_handler = openstack_driver.NeutronClient(self.config)
        port_details = super(NFPGBPNetworkDriver, self).get_port_details(
            token, _port_id)
        # self.network_handler = openstack_driver.GBPClient(self.config)
        return port_details

    def get_networks(self, token, filters):
        with nfp_ctx_mgr.GBPContextManager as gcm:
            return gcm.retry(
                self.network_handler.get_policy_target_groups,
                token, filters=filters)

    def set_promiscuos_mode(self, token, port_id, enable_port_security):
        port_id = self.get_port_id(token, port_id)
        # self.network_handler = openstack_driver.NeutronClient(self.config)
        super(NFPGBPNetworkDriver, self).set_promiscuos_mode(
            token, port_id, enable_port_security)
        # self.network_handler = openstack_driver.GBPClient(self.config)

    def set_promiscuos_mode_fast(self, token, port_id, enable_port_security):
        # self.network_handler = openstack_driver.NeutronClient(self.config)
        super(NFPGBPNetworkDriver, self).set_promiscuos_mode(
            token, port_id, enable_port_security)
        # self.network_handler = openstack_driver.GBPClient(self.config)

    def get_service_profile(self, token, service_profile_id):
        with nfp_ctx_mgr.GBPContextManager as gcm:
            return gcm.retry(self.network_handler.get_service_profile, token,
                             service_profile_id)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import commands
import logging
import sys

from libs import config_libs
from libs import utils_libs
from libs import verify_libs


def main():
    # Run the Testcases:
    test = test_gbp_prs_pr_shared_func()
    if test.test_gbp_prs_pr_shared_func_1() == 0:
        test.cleanup(tc_name='TESTCASE_GBP_PRS_PR_SHARED_INTEG_1')
    if test.test_gbp_prs_pr_shared_func_2() == 0:
        test.cleanup(tc_name='TESTCASE_GBP_PRS_PR_SHARED_INTEG_2')
    test.cleanup()
    utils_libs.report_results('test_gbp_prs_pr_shared_func',
                              'test_results_admin.txt')
    sys.exit(1)


class test_gbp_prs_pr_shared_func(object):

    # Initialize logging
    logging.basicConfig(
        format='%(asctime)s [%(levelname)s] %(name)s - %(message)s',
        level=logging.WARNING)
    _log = logging.getLogger(__name__)
    cmd = 'rm /tmp/test_gbp_prs_pr_shared_func.log'
    commands.getoutput(cmd)
    hdlr = logging.FileHandler('/tmp/test_gbp_prs_pr_shared_func.log')
    formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
    hdlr.setFormatter(formatter)
    _log.addHandler(hdlr)
    _log.setLevel(logging.INFO)
    _log.setLevel(logging.DEBUG)

    def __init__(self):
        """
        Init def
        """
        self._log.info(
            "\n## START OF GBP POLICY_RULE_SET FUNCTIONALITY TESTSUITE\n")
        self.gbpcfg = config_libs.Gbp_Config()
        self.gbpverify = verify_libs.Gbp_Verify()
        self.act_name = 'demo_pa'
        self.cls_name = 'demo_pc'
        self.rule_name = 'demo_pr'
        self.ruleset_name = 'demo_prs'

    def cleanup(self, tc_name=''):
        if tc_name != '':
            self._log.info('%s: FAILED' % (tc_name))
        for obj in ['ruleset', 'rule', 'classifier', 'action']:
            self.gbpcfg.gbp_del_all_anyobj(obj)

    def test_gbp_prs_pr_shared_func_1(self):

        self._log.info(
            "\n############################################################\n"
            "TESTCASE_GBP_PRS_PR_SHARED_INTEG_1: TO "
            "CREATE/UPDATE/VERIFY/DELETE/ ASSOCIATING MULTIPLE PRs to 1 "
            "POLICY RULESET \n"
            "TEST_STEP::\n"
            "Create Multiple Policy Rules with shared=True\n"
            "Create Policy RuleSet by associating all the Policy "
            "Rules and shared=False(default)\n"
            "Verify that multiple Policy Rules are associated to "
            "the Policy RuleSet\n"
            "Update the Policy RuleSet with shared=True\n"
            "Verify the Policy RuleSet's shared=True\n"
            "##############################################################\n")

        # Testcase work-flow starts
        self._log.info(
            '\n## Step 1A: Create new PA ,new PC, 4 PRs using the '
            'same PA & PC##\n')
        new_cls_uuid = self.gbpcfg.gbp_policy_cfg_all(
            1, 'classifier', 'grppol_pc1', shared='True')
        if new_cls_uuid == 0:
            self._log.info(
                "\nNew Classifier Create Failed, hence "
                "Testcase_gbp_prs_pr_shared_integ_1 ABORTED\n")
            return 0
        new_act_uuid = self.gbpcfg.gbp_policy_cfg_all(
            1, 'action', 'grppol_pa1', shared='True')
        if new_act_uuid == 0:
            self._log.info(
                "\nNew Action Create Failed, hence "
                "testcase_gbp_prs_pr_shared_integ_1 ABORTED\n")
            return 0
        rule_uuid_list = []
        for i in range(4):
            new_rule_uuid = self.gbpcfg.gbp_policy_cfg_all(
                1,
                'rule',
                'grppol_pr_%s' %
                (i),
                classifier=new_cls_uuid,
                action=new_act_uuid,
                description="'For devstack demo'",
                shared="True")
            if new_rule_uuid == 0:
                self._log.info(
                    "\nNew Rule Create Failed, hence "
                    "testcase_gbp_prs_pr_shared_integ_1 ABORTED\n")
                return 0
            rule_uuid_list.append(new_rule_uuid)
        ruleset_uuid = self.gbpcfg.gbp_policy_cfg_all(
            1,
            'ruleset',
            'grppol_prs_many',
            policy_rule='"%s %s %s %s"' %
            (rule_uuid_list[0],
             rule_uuid_list[1],
             rule_uuid_list[2],
             rule_uuid_list[3]),
            description="'For devstack demo'",
            shared='False')
        if ruleset_uuid == 0:
            self._log.info(
                "\nStep 2: Creating Policy RuleSet with multiple "
                "PRs(shared=True) and shared=False , Failed")
            return 0
        # Verify starts
        if self.gbpverify.gbp_policy_verify_all(
                1,
                'ruleset',
                ruleset_uuid,
                name='grppol_prs_many',
                description='For devstack demo',
                shared='False') == 0:
            self._log.info(
                "# Step 2B: Verify Policy RuleSet and its "
                "shared='False' == Failed")
            return 0
        if self.gbpverify.gbp_obj_ver_attr_all_values(
                'ruleset', 'grppol_prs_many', 'policy_rules',
                rule_uuid_list) == 0:
            self._log.info(
                "# Step 2C: Verify Policy RuleSet and its "
                "Multiple PRs using -show option == Failed")
            return 0
        # Update the Policy RuleSet with shared=True and
        # update MUST fail as it is not supported by RMD
        if self.gbpcfg.gbp_policy_cfg_all(
                2, 'ruleset', 'grppol_prs_many', shared='True') != 0:
            self._log.info(
                "# Step 3: Updating Policy RuleSet's Attribute "
                "shared=True DID NOT Fail")
            return 0
        if self.gbpverify.gbp_obj_ver_attr_all_values(
                'ruleset', 'grppol_prs_many', 'policy_rules',
                rule_uuid_list) == 0:
            self._log.info(
                "# Step 3A: Verify Policy RuleSet and its "
                "Multiple PRs, == Failed")
            return 0
        if self.gbpverify.gbp_policy_verify_all(
                1,
                'ruleset',
                ruleset_uuid,
                name='grppol_prs_many',
                description='For devstack demo',
                shared='False') == 0:
            self._log.info(
                "# Step 3B: Verify Policy RuleSet and its "
                "shared=False, == Failed")
            return 0
        self._log.info("\nTESTCASE_GBP_PRS_PR_SHARED_INTEG_1: PASSED")
        self.cleanup()
        return 1

    def test_gbp_prs_pr_shared_func_2(self):
        self._log.info(
            "\n############################################################\n"
            "TESTCASE_GBP_PRS_PR_SHARED_INTEG_2: TO "
            "CREATE/UPDATE/VERIFY/DELETE/ ASSOCIATING MULTIPLE "
            "PRs to 1 POLICY RULESET\n"
            "TEST_STEP::\n"
            "Create Multiple Policy Rules witha mix of shared=True "
            "and shared=False\n"
            "Create Policy RuleSet by associating all the Policy "
            "Rules and shared=False(default)\n"
            "Verify that multiple Policy Rules are associated to "
            "the Policy RuleSet\n"
            "Update the Policy RuleSet with shared=True and update "
            "should fail\n"
            "Verify the Policy RuleSet's continues with attribute "
            "shared=False\n"
            "##############################################################\n")

        # Testcase work-flow starts
        self._log.info(
            '\n## Step 1A: Create new PA ,new PC, 4 PRs using the '
            'same PA & PC##\n')
        new_cls_uuid = self.gbpcfg.gbp_policy_cfg_all(
            1, 'classifier', 'grppol_pc1', shared='True')
        if new_cls_uuid == 0:
            self._log.info(
                "\nNew Classifier Create Failed, hence "
                "testcase_gbp_prs_pr_shared_integ_2 ABORTED\n")
            return 0
        new_act_uuid = self.gbpcfg.gbp_policy_cfg_all(
            1, 'action', 'grppol_pa1', shared='True')
        if new_act_uuid == 0:
            self._log.info(
                "\nNew Action Create Failed, hence "
                "testcase_gbp_prs_pr_shared_integ_2 ABORTED\n")
            return 0
        rule_uuid_list = []
        shared_flag = ['True', 'False', 'True', 'False']
        for i in range(4):
            new_rule_uuid = self.gbpcfg.gbp_policy_cfg_all(
                1,
                'rule',
                'grppol_pr_%s' %
                (i),
                classifier=new_cls_uuid,
                action=new_act_uuid,
                description="'For devstack demo'",
                shared=shared_flag[i])
            if new_rule_uuid == 0:
                self._log.info(
                    "\nNew Rule Create Failed, hence "
                    "testcase_gbp_prs_pr_shared_integ_2 ABORTED\n")
                return 0
            rule_uuid_list.append(new_rule_uuid)
        ruleset_uuid = self.gbpcfg.gbp_policy_cfg_all(
            1,
            'ruleset',
            'grppol_prs_many',
            policy_rule='"%s %s %s %s"' %
            (rule_uuid_list[0],
             rule_uuid_list[1],
             rule_uuid_list[2],
             rule_uuid_list[3]),
            description="'For devstack demo'",
            shared='False')
        if ruleset_uuid == 0:
            self._log.info(
                "\nStep 2: Creating Policy RuleSet with multiple "
                "PRs(shared=True) and shared=False , Failed")
            return 0
        # Verify starts
        if self.gbpverify.gbp_policy_verify_all(
                1,
                'ruleset',
                ruleset_uuid,
                name='grppol_prs_many',
                description='For devstack demo',
                shared='False') == 0:
            self._log.info(
                "# Step 2B: Verify Policy RuleSet and its "
                "shared='False' == Failed")
            return 0
        if self.gbpverify.gbp_obj_ver_attr_all_values(
                'ruleset', 'grppol_prs_many', 'policy_rules',
                rule_uuid_list) == 0:
            self._log.info(
                "# Step 2C: Verify Policy RuleSet and its "
                "Multiple PRs using -show option == Failed")
            return 0
        # Update and Verify the PRS by updating the PRs(removing few existing
        # ones)
        if self.gbpcfg.gbp_policy_cfg_all(
                2, 'ruleset', 'grppol_prs_many', shared='True') != 0:
            self._log.info(
                "# Step 3: Updating Policy RuleSet's"
                " Attribute shared=True did NOT fail")
            return 0
        if self.gbpverify.gbp_obj_ver_attr_all_values(
                'ruleset', 'grppol_prs_many', 'policy_rules',
                rule_uuid_list) == 0:
            self._log.info(
                "# Step 3A: Verify Policy RuleSet and its "
                "Multiple PRs using -show option == Failed")
            return 0
        if self.gbpverify.gbp_policy_verify_all(
                1,
                'ruleset',
                ruleset_uuid,
                name='grppol_prs_many',
                description='For devstack demo',
                shared='False') == 0:
            self._log.info(
                "# Step 3B: Verify Policy RuleSet and its "
                "shared=False, == Failed")
            return 0

        self._log.info("\nTESTCASE_GBP_PRS_PR_SHARED_INTEG_2: PASSED")
        return 1


if __name__ == '__main__':
    main()



# Copyright (c) 2016 Cisco Systems Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.tests.unit.plugins.ml2.drivers import (
    mechanism_logger as ml2_logger)
from oslo_log import log

from gbpservice.neutron.plugins.ml2plus import driver_api

LOG = log.getLogger(__name__)


class LoggerPlusMechanismDriver(driver_api.MechanismDriver,
                                ml2_logger.LoggerMechanismDriver):
    """Mechanism driver that logs all calls and parameters made.

    Generally used for testing and debugging.
    """

    def initialize(self):
        LOG.info("initialize called")

    def start_rpc_listeners(self):
        LOG.info("start_rpc_listeners called")
        return []

    def ensure_tenant(self, plugin_context, tenant_id):
        LOG.info("ensure_tenant called with tenant_id %s", tenant_id)

    def _log_subnetpool_call(self, method_name, context):
        LOG.info("%(method)s called with subnetpool settings %(current)s "
                 "(original settings %(original)s)",
                 {'method': method_name,
                  'current': context.current,
                  'original': context.original})

    def create_subnetpool_precommit(self, context):
        self._log_subnetpool_call("create_subnetpool_precommit",
                                  context)

    def create_subnetpool_postcommit(self, context):
        self._log_subnetpool_call("create_subnetpool_postcommit",
                                  context)

    def update_subnetpool_precommit(self, context):
        self._log_subnetpool_call("update_subnetpool_precommit",
                                  context)

    def update_subnetpool_postcommit(self, context):
        self._log_subnetpool_call("update_subnetpool_postcommit",
                                  context)

    def delete_subnetpool_precommit(self, context):
        self._log_subnetpool_call("delete_subnetpool_precommit",
                                  context)

    def delete_subnetpool_postcommit(self, context):
        self._log_subnetpool_call("delete_subnetpool_postcommit",
                                  context)

    def _log_address_scope_call(self, method_name, context):
        LOG.info("%(method)s called with address_scope settings %(current)s "
                 "(original settings %(original)s)",
                 {'method': method_name,
                  'current': context.current,
                  'original': context.original})

    def create_address_scope_precommit(self, context):
        self._log_address_scope_call("create_address_scope_precommit",
                                     context)

    def create_address_scope_postcommit(self, context):
        self._log_address_scope_call("create_address_scope_postcommit",
                                     context)

    def update_address_scope_precommit(self, context):
        self._log_address_scope_call("update_address_scope_precommit",
                                     context)

    def update_address_scope_postcommit(self, context):
        self._log_address_scope_call("update_address_scope_postcommit",
                                     context)

    def delete_address_scope_precommit(self, context):
        self._log_address_scope_call("delete_address_scope_precommit",
                                     context)

    def delete_address_scope_postcommit(self, context):
        self._log_address_scope_call("delete_address_scope_postcommit",
                                     context)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from networking_sfc.db import flowclassifier_db as flc_db
from networking_sfc.extensions import flowclassifier
from networking_sfc.services.flowclassifier.drivers import base
from neutron_lib.api import validators
from neutron_lib.callbacks import events
from neutron_lib.callbacks import registry
from neutron_lib.callbacks import resources
from neutron_lib.plugins import directory
from oslo_log import log as logging
import sqlalchemy as sa
from sqlalchemy.ext import baked

from gbpservice.neutron.plugins.ml2plus.drivers.apic_aim import constants
from gbpservice.neutron.services.grouppolicy.common import exceptions as exc
from gbpservice.neutron.services.sfc.aim import constants as sfc_cts
from gbpservice.neutron.services.sfc.aim import exceptions as sfc_exc

LOG = logging.getLogger(__name__)
flowclassifier.SUPPORTED_L7_PARAMETERS.update(sfc_cts.AIM_FLC_L7_PARAMS)

BAKERY = baked.bakery(_size_alert=lambda c: LOG.warning(
    "sqlalchemy baked query cache size exceeded in %s", __name__))


class FlowclassifierAIMDriverBase(base.FlowClassifierDriverBase):
    def create_flow_classifier_precommit(self, context):
        pass

    def create_flow_classifier(self, context):
        pass

    def update_flow_classifier(self, context):
        pass

    def delete_flow_classifier(self, context):
        pass


@registry.has_registry_receivers
class FlowclassifierAIMDriver(FlowclassifierAIMDriverBase):
    """SFC Driver mapping for AIM."""

    def initialize(self):
        self._core_plugin = None

    @property
    def plugin(self):
        if not self._core_plugin:
            self._core_plugin = directory.get_plugin()
            if not self._core_plugin:
                LOG.error("No Core plugin found.")
                raise exc.GroupPolicyDeploymentError()
        return self._core_plugin

    def create_flow_classifier_precommit(self, context):
        self._validate_flow_classifier(context)
        registry.notify(constants.GBP_FLOW_CLASSIFIER, events.PRECOMMIT_CREATE,
                        self, driver_context=context)

    def update_flow_classifier_precommit(self, context):
        self._validate_flow_classifier(context)
        registry.notify(constants.GBP_FLOW_CLASSIFIER, events.PRECOMMIT_UPDATE,
                        self, driver_context=context)

    def delete_flow_classifier_precommit(self, context):
        registry.notify(constants.GBP_FLOW_CLASSIFIER, events.PRECOMMIT_DELETE,
                        self, driver_context=context)

    def _validate_flow_classifier(self, context):
        fc = context.current
        # Verify L7 params are set
        l7_p = fc['l7_parameters']
        if any(x for x in sfc_cts.AIM_FLC_L7_PARAMS.keys()
               if not validators.is_attr_set(l7_p.get(x))):
            raise sfc_exc.BadFlowClassifier(
                params=sfc_cts.AIM_FLC_L7_PARAMS.keys())
        # Verify standard params are set
        # TODO(ivar): src and dst prefix are needed only for SVI networks
        if any(x for x in sfc_cts.AIM_FLC_PARAMS
               if not validators.is_attr_set(fc.get(x))):
            raise sfc_exc.BadFlowClassifier(params=sfc_cts.AIM_FLC_PARAMS)
        # Verify networks exist
        src_net = self.plugin.get_network(
            context._plugin_context, l7_p[sfc_cts.LOGICAL_SRC_NET])
        if l7_p[sfc_cts.LOGICAL_SRC_NET] != l7_p[sfc_cts.LOGICAL_DST_NET]:
            # Verify dst existence
            self.plugin.get_network(context._plugin_context,
                                    l7_p[sfc_cts.LOGICAL_DST_NET])
        elif src_net.get('apic:svi') is False:
            # Same network, not SVI
            raise sfc_exc.FlowClassifierSameSrcDstNetworks()

        if validators.is_attr_set(fc.get('source_ip_prefix')) and (
                fc.get('source_ip_prefix') == fc.get('destination_ip_prefix')):
            # Same subnet for source and dst is not allowed. For overlapping
            # (but not same) subnets LPM will be applied.
            raise sfc_exc.FlowClassifierSameSubnet()

        # TODO(ivar): Any other parameter is unsupported, for now just
        # unenforced.

        # TODO(ivar): if source and destination ports are to a private network
        # source/destination CIDRs are not required

        # TODO(ivar): only one classifier can be provider (destination) if
        # the network is not SVI.

    def _get_classifiers_by_network_id(self, plugin_context, network_id):
        context = plugin_context
        with context.session.begin(subtransactions=True):
            classifier_ids = []
            for keyword in [sfc_cts.LOGICAL_SRC_NET, sfc_cts.LOGICAL_DST_NET]:
                query = BAKERY(lambda s: s.query(
                    flc_db.L7Parameter))
                query += lambda q: q.filter_by(
                    keyword=sa.bindparam('keyword'))
                query += lambda q: q.filter_by(
                    value=sa.bindparam('network_id'))
                classifier_ids.extend(
                    [x.classifier_id for x in query(context.session).params(
                        keyword=keyword, network_id=network_id).all()])

            return classifier_ids

    @registry.receives(resources.NETWORK, [events.PRECOMMIT_DELETE])
    def _handle_network_delete(self, rtype, event, trigger, context,
                               network_id, **kwargs):
        flc_ids = self._get_classifiers_by_network_id(context, network_id)
        if flc_ids:
            # TODO(ivar): instead of raising, we could try deleting the flow
            # classifier, which would fail (and rollback the transaction) if
            # in use.
            raise sfc_exc.NetworkInUseByFlowClassifiers(ids=flc_ids)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ast
import zlib

from neutron.tests import base
import oslo_serialization.jsonutils as jsonutils
import pecan
from pecan import rest
import webtest

from gbpservice.nfp.pecan import constants

import six
if six.PY3:
    from importlib import reload

setattr(pecan, 'mode', constants.base)


ERROR = 'error'
UNHANDLED = 'unhandled'
FAILURE = 'failure'


class ControllerTestCase(base.BaseTestCase, rest.RestController):

    """This class contains  unittest cases for REST server of configurator.

    This class tests success and failure cases for all the HTTP requests which
    are implemented in REST server.

    """

    def setUp(self):
        """Standard method of TestCase to setup environment before each test.

        This method set the value of required variables that is used in
        test cases before execution of each test case.
        """
        super(ControllerTestCase, self).setUp()
        from gbpservice.nfp.pecan.api import root_controller
        reload(root_controller)
        RootController = root_controller.RootController()
        self.app = webtest.TestApp(pecan.make_app(RootController))
        self.data = {'info': {'service_type': 'firewall',
                              'service_vendor': '',
                              'context': {'foo': 'foo'}},
                     'config': [{'resource': 'heat',
                                 'resource_data': {'some_data': 'some_value'}}]
                     }
        self.data_non_heat = {'info': {'service_type': 'firewall',
                                       'service_vendor': '',
                                       'context': {'foo': 'foo'}},
                              'config': [{'resource': 'non-heat',
                                          'resource_data': {
                                                'some_data': 'some_value'}}]
                              }

    def post_create_network_function_config_with_heat(self,
            operation='create'):
        """Tests HTTP post request create_network_function_device_config.

        Returns: none

        """

        self.data['info']['context']['operation'] = operation
        response = self.app.post(
                '/v1/nfp/create_network_function_config',
                zlib.compress(jsonutils.dumps(self.data)),
                content_type='application/octet-stream')
        self.assertEqual(response.status_code, 200)

    def post_create_network_function_config_with_others(self,
            operation='create'):
        """Tests HTTP post request create_network_function_device_config.

        Returns: none

        """

        self.data_non_heat['info']['context']['operation'] = operation
        response = self.app.post(
                '/v1/nfp/create_network_function_config',
                zlib.compress(jsonutils.dumps(self.data_non_heat)),
                content_type='application/octet-stream')
        self.assertEqual(response.status_code, 200)

    def post_delete_network_function_config_with_heat(self,
            operation='delete'):
        """Tests HTTP post request delete_network_function_device_config.

        Returns: none

        """

        response = self.app.post(
                '/v1/nfp/delete_network_function_config',
                zlib.compress(jsonutils.dumps(self.data)),
                content_type='application/octet-stream')
        self.assertEqual(response.status_code, 200)

    def post_delete_network_function_config_with_others(self,
            operation='delete'):
        """Tests HTTP post request delete_network_function_device_config.

        Returns: none

        """

        response = self.app.post(
                '/v1/nfp/delete_network_function_config',
                zlib.compress(jsonutils.dumps(self.data_non_heat)),
                content_type='application/octet-stream')
        self.assertEqual(response.status_code, 200)

    def test_get_notifications(self):
        """Tests HTTP get request get_notifications.

        Returns: none

        """
        config_data = self.data['config'][0]
        info_data = self.data['info']
        service_type = info_data['service_type']
        notification_context = self.data['info']['context']
        resource = config_data['resource']
        response_unhandled = {'info': {'service_type': service_type,
                                       'context': notification_context},
                              'notification': [{
                                        'resource': resource,
                                        'data': {
                                            'status_code': UNHANDLED}}]
                              }
        response_error = {'info': {'service_type': service_type,
                                   'context': notification_context},
                          'notification': [{
                                    'resource': 'non-heat',
                                    'data': {
                                        'status_code': FAILURE,
                                        'error_msg': (
                                            'Unsupported resource type')}}]
                          }
        self.post_create_network_function_config_with_heat(operation='create')
        self.post_delete_network_function_config_with_heat(operation='delete')
        self.post_create_network_function_config_with_others(
                operation='create')
        self.post_delete_network_function_config_with_others(
                operation='delete')
        response = self.app.get(
                '/v1/nfp/get_notifications')
        response_str = zlib.decompress(response.body)
        response_expected = ast.literal_eval(response_str)
        self.assertEqual(response_expected[0], response_unhandled)
        self.assertEqual(response_expected[1], response_unhandled)
        self.assertEqual(response_expected[2], response_error)
        self.assertEqual(response_expected[3], response_error)
        self.assertEqual(response.status_code, 200)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# The following two are required when running tox since the nsx packages
# have a foreign key dependency on LBaaS tables. Those tables are not
# getting loaded without these explicit imports.
import sys

from neutron.agent import securitygroups_rpc
from neutron.api import extensions
from neutron.quota import resource
from neutron.quota import resource_registry
from neutron.scheduler import l3_agent_scheduler
from stevedore import named

from gbpservice.network.neutronv2 import local_api


# The following is to avoid excessive logging in the UTs
extensions.LOG.warning = extensions.LOG.info
resource_registry.LOG.warning = resource_registry.LOG.info
l3_agent_scheduler.LOG.warning = l3_agent_scheduler.LOG.info
securitygroups_rpc.LOG.warning = securitygroups_rpc.LOG.info
local_api.LOG.warning = local_api.LOG.info
named.LOG.warning = named.LOG.info


orig_warning = resource.LOG.warning


def warning(*args):
    try:
        for val in sys._getframe(1).f_locals.values():
            if isinstance(val, resource.TrackedResource) and (
                sys._getframe(1).f_code.co_name == (
                    'unregister_events')):
                return
    except Exception:
        pass
    orig_warning(*args)


resource.LOG.warning = warning



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

import six

from gbpservice.contrib.nfp.configurator.agents import agent_base
from gbpservice.contrib.nfp.configurator.lib import data_filter
from gbpservice.contrib.nfp.configurator.lib import lbv2_constants as lb_const
from gbpservice.contrib.nfp.configurator.lib import utils
from gbpservice.nfp.common import exceptions
from gbpservice.nfp.core import event as nfp_event
from gbpservice.nfp.core import log as nfp_logging
from gbpservice.nfp.core import module as nfp_api

LOG = nfp_logging.getLogger(__name__)

""" Implements LBaaS response path to Neutron plugin.
Methods of this class are invoked by the LBaaSV2EventHandler class and also
by driver class for sending response from driver to the LBaaS Neutron plugin.
"""


class LBaaSV2RpcSender(data_filter.Filter):

    def __init__(self, sc):
        self.notify = agent_base.AgentBaseNotification(sc)

    def update_status(self, obj_type, obj_id, root_lb_id,
                      provisioning_status, operating_status,
                      agent_info, obj=None):
        """ Enqueues the response from LBaaS V2 operation to neutron plugin.

        :param obj_type: object type
        :param obj_id: object id
        :param root_lb_id: root loadbalancer id
        :param provisioning_status: an enum of ('ACTIVE', 'PENDING_CREATE',
        'PENDING_UPDATE', 'PENDING_DELETE', 'ERROR')
        :param operating_status:  an enum of
        ('ONLINE', 'OFFLINE', 'DEGRADED', 'ERROR')

        """

        msg = {'info': {'service_type': lb_const.SERVICE_TYPE,
                        'context': agent_info['context']},
               'notification': [{'resource': agent_info['resource'],
                                 'data':{'obj_type': obj_type,
                                         'obj_id': obj_id,
                                         'notification_type': 'update_status',
                                         'root_lb_id': root_lb_id,
                                         'provisioning_status':
                                             provisioning_status,
                                         'operating_status':
                                             operating_status,
                                         obj_type: obj}}]
               }
        LOG.info("Sending Notification 'Update Status' "
                 "for resource: %(resource)s with Provisioning status:"
                 "%(p_status)s and Operating status:%(o_status)s",
                 {'resource': agent_info['resource'],
                  'p_status': provisioning_status,
                  'o_status': operating_status})
        self.notify._notification(msg)

    # REVISIT(jiahao): need to revisit how lbaasv2 update lb stats,
    # will add in visibility patch
    def update_pool_stats(self, pool_id, stats, context, pool=None):
        """ Enqueues the response from LBaaS operation to neutron plugin.

        :param pool_id: pool id
        :param stats: statistics of that pool

        """
        msg = {'info': {'service_type': lb_const.SERVICE_TYPE,
                        'context': context.to_dict()},
               'notification': [{'resource': 'pool',
                                 'data': {'pool_id': pool_id,
                                          'stats': stats,
                                          'notification_type': (
                                              'update_pool_stats'),
                                          'pool': pool_id}}]
               }
        LOG.info("Sending Notification 'Update Pool Stats' "
                 "for pool: %(pool_id)s with stats:%(stats)s",
                 {'pool_id': pool_id,
                  'stats': stats})
        self.notify._notification(msg)


"""Implements APIs invoked by configurator for processing RPC messages.

RPC client of configurator module receives RPC messages from REST server
and invokes the API of this class. The instance of this class is registered
with configurator module using register_service_agent API. Configurator module
identifies the service agent object based on service type and invokes ones of
the methods of this class to configure the device.

"""


class LBaaSv2RpcManager(agent_base.AgentBaseRPCManager):

    def __init__(self, sc, conf):
        """Instantiates child and parent class objects.

        :param sc: Service Controller object that is used for interfacing
        with core service controller.
        :param conf: Configuration object that is used for configuration
        parameter access.

        """

        super(LBaaSv2RpcManager, self).__init__(sc, conf)

    def _send_event(self, event_id, data, serialize=False, binding_key=None,
                    key=None):
        """Posts an event to framework.

        :param event_id: Unique identifier for the event
        :param event_key: Event key for serialization
        :param serialize: Serialize the event
        :param binding_key: binding key to be used for serialization
        :param key: event key

        """

        ev = self.sc.new_event(id=event_id, data=data)
        ev.key = key
        ev.sequence = serialize
        ev.binding_key = binding_key
        self.sc.post_event(ev)

    def create_loadbalancer(self, context, loadbalancer, driver_name):
        """Enqueues event for worker to process create loadbalancer request.

        :param context: RPC context
        :param loadbalancer: loadbalancer resource to be created

        Returns: None

        """
        LOG.info("Received request 'Create Loadbalancer' for LB:%(lb)s "
                 "with driver:%(driver_name)s",
                 {'lb': loadbalancer['id'],
                  'driver_name': driver_name})
        arg_dict = {'context': context,
                    lb_const.LOADBALANCER: loadbalancer,
                    'driver_name': driver_name
                    }
        self._send_event(lb_const.EVENT_CREATE_LOADBALANCER_V2, arg_dict,
                         serialize=True, binding_key=loadbalancer['id'],
                         key=loadbalancer['id'])

    def update_loadbalancer(self, context, old_loadbalancer, loadbalancer):
        """Enqueues event for worker to process update loadbalancer request.

        :param context: RPC context
        :param old_loadbalancer: old loadbalancer resource to be updated
        :param loadbalancer: new loadbalancer resource

        Returns: None

        """
        old_val, new_val = self.get_diff_of_dict(
            old_loadbalancer, loadbalancer)
        arg_dict = {'context': context,
                    lb_const.OLD_LOADBALANCER: old_loadbalancer,
                    lb_const.LOADBALANCER: loadbalancer,
                    }
        LOG.info("Received request 'Update Loadbalancer' for LB:%(lb)s "
                 "with new Param:%(new_val)s and old Param:%(old_val)s",
                 {'lb': loadbalancer['id'],
                  'new_val': new_val,
                  'old_val': old_val})
        self._send_event(lb_const.EVENT_UPDATE_LOADBALANCER_V2, arg_dict,
                         serialize=True, binding_key=loadbalancer['id'],
                         key=loadbalancer['id'])

    def delete_loadbalancer(self, context, loadbalancer):
        """Enqueues event for worker to process delete loadbalancer request.

        :param context: RPC context
        :param loadbalancer: loadbalancer resource to be deleted

        Returns: None

        """
        LOG.info("Received request 'Delete Loadbalancer' for LB:%(lb)s ",
                 {'lb': loadbalancer['id']})

        arg_dict = {'context': context,
                    lb_const.LOADBALANCER: loadbalancer,
                    }
        self._send_event(lb_const.EVENT_DELETE_LOADBALANCER_V2, arg_dict,
                         serialize=True, binding_key=loadbalancer['id'],
                         key=loadbalancer['id'])

    def create_listener(self, context, listener):
        """Enqueues event for worker to process create listener request.

        :param context: RPC context
        :param listener: listener resource to be created

        Returns: None

        """
        LOG.info("Received request 'Create Listener' for LB:%(lb)s ",
                 {'lb': listener['loadbalancer_id']})
        arg_dict = {'context': context,
                    lb_const.LISTENER: listener,
                    }
        self._send_event(lb_const.EVENT_CREATE_LISTENER_V2, arg_dict,
                         serialize=True,
                         binding_key=listener['loadbalancer_id'],
                         key=listener['id'])

    def update_listener(self, context, old_listener, listener):
        """Enqueues event for worker to process update listener request.

        :param context: RPC context
        :param old_listener: old listener resource to be updated
        :param listener: new listener resource

        Returns: None

        """
        old_val, new_val = self.get_diff_of_dict(old_listener, listener)
        LOG.info("Received request 'Update Listener' for Listener:"
                 "%(listener)s in LB:%(lb_id)s with new Param:"
                 "%(new_val)s and old Param:%(old_val)s",
                 {'lb_id': listener['loadbalancer_id'],
                  'listener': listener['id'],
                  'old_val': old_val,
                  'new_val': new_val})
        arg_dict = {'context': context,
                    lb_const.OLD_LISTENER: old_listener,
                    lb_const.LISTENER: listener,
                    }
        self._send_event(lb_const.EVENT_UPDATE_LISTENER_V2, arg_dict,
                         serialize=True,
                         binding_key=listener['loadbalancer_id'],
                         key=listener['id'])

    def delete_listener(self, context, listener):
        """Enqueues event for worker to process delete listener request.

        :param context: RPC context
        :param listener: listener resource to be deleted

        Returns: None

        """
        LOG.info("Received request 'Delete Listener' for LB:%(lb)s ",
                 {'lb': listener['loadbalancer_id']})
        arg_dict = {'context': context,
                    lb_const.LISTENER: listener,
                    }
        self._send_event(lb_const.EVENT_DELETE_LISTENER_V2, arg_dict,
                         serialize=True,
                         binding_key=listener['loadbalancer_id'],
                         key=listener['id'])

    def create_pool(self, context, pool):
        """Enqueues event for worker to process create pool request.

        :param context: RPC context
        :param pool: pool resource to be created

        Returns: None

        """
        LOG.info("Received request 'Create Pool' for Pool:%(pool_id)s ",
                 {'pool_id': pool['id']})
        arg_dict = {'context': context,
                    lb_const.POOL: pool
                    }
        # REVISIT(jiahao) M:N pool is not yet implemented.
        self._send_event(lb_const.EVENT_CREATE_POOL_V2, arg_dict,
                         serialize=True,
                         binding_key=pool['loadbalancer_id'],
                         key=pool['id'])

    def update_pool(self, context, old_pool, pool):
        """Enqueues event for worker to process update pool request.

        :param context: RPC context
        :param old_pool: old pool resource to be updated
        :param pool: new pool resource

        Returns: None

        """
        old_val, new_val = self.get_diff_of_dict(old_pool, pool)
        LOG.info("Received request 'Update Pool' for Pool:%(pool)s "
                 "in LB:%(lb_id)s with new Param:%(new_val)s and "
                 "old Param:%(old_val)s",
                 {'pool': pool['id'],
                  'lb_id': pool['loadbalancer_id'],
                  'old_val': old_val,
                  'new_val': new_val})
        arg_dict = {'context': context,
                    lb_const.OLD_POOL: old_pool,
                    lb_const.POOL: pool,
                    }
        self._send_event(lb_const.EVENT_UPDATE_POOL_V2, arg_dict,
                         serialize=True,
                         binding_key=pool['loadbalancer_id'],
                         key=pool['id'])

    def delete_pool(self, context, pool):
        """Enqueues event for worker to process delete pool request.

        :param context: RPC context
        :param pool: pool resource to be deleted

        Returns: None

        """
        LOG.info("Received request 'Delete Pool' for Pool:%(pool_id)s ",
                 {'pool_id': pool['id']})
        arg_dict = {'context': context,
                    lb_const.POOL: pool,
                    }
        self._send_event(lb_const.EVENT_DELETE_POOL_V2, arg_dict,
                         serialize=True,
                         binding_key=pool['loadbalancer_id'],
                         key=pool['id'])

    def create_member(self, context, member):
        """Enqueues event for worker to process create member request.

        :param context: RPC context
        :param member: member resource to be created

        Returns: None

        """
        LOG.info("Received request 'Create Member' for Pool:%(pool_id)s ",
                 {'pool_id': member['pool_id']})
        arg_dict = {'context': context,
                    lb_const.MEMBER: member,
                    }
        self._send_event(lb_const.EVENT_CREATE_MEMBER_V2, arg_dict,
                         serialize=True,
                         binding_key=member[lb_const.POOL]['loadbalancer_id'],
                         key=member['id'])

    def update_member(self, context, old_member, member):
        """Enqueues event for worker to process update member request.

        :param context: RPC context
        :param old_member: old member resource to be updated
        :param member: new member resource

        Returns: None

        """
        old_val, new_val = self.get_diff_of_dict(old_member, member)
        LOG.info("Received request 'Update Member' for Member:"
                 "%(member_id)s in Pool:%(pool_id)s with new Param:"
                 "%(new_val)s and old Param:%(old_val)s",
                 {'pool_id': member['pool_id'],
                  'member_id': member['id'],
                  'old_val': old_val,
                  'new_val': new_val})
        arg_dict = {'context': context,
                    lb_const.OLD_MEMBER: old_member,
                    lb_const.MEMBER: member,
                    }
        self._send_event(lb_const.EVENT_UPDATE_MEMBER_V2, arg_dict,
                         serialize=True,
                         binding_key=member[lb_const.POOL]['loadbalancer_id'],
                         key=member['id'])

    def delete_member(self, context, member):
        """Enqueues event for worker to process delete member request.

        :param context: RPC context
        :param member: member resource to be deleted

        Returns: None

        """
        LOG.info("Received request 'Delete Member' for Pool:"
                 "%(pool_id)s ",
                 {'pool_id': member['pool_id']})
        arg_dict = {'context': context,
                    lb_const.MEMBER: member,
                    }
        self._send_event(lb_const.EVENT_DELETE_MEMBER_V2, arg_dict,
                         serialize=True,
                         binding_key=member[lb_const.POOL]['loadbalancer_id'],
                         key=member['id'])

    def create_healthmonitor(self, context, healthmonitor):
        """Enqueues event for worker to process create health monitor request.

        :param context: RPC context
        :param health_monitor: health_monitor resource to be created
        :param pool_id: pool_id to which health monitor is associated

        Returns: None

        """
        LOG.info("Received request 'Create Pool Health Monitor' for"
                 "Health monitor:%(hm)s",
                 {'hm': healthmonitor['id']})
        arg_dict = {'context': context,
                    lb_const.HEALTHMONITOR: healthmonitor
                    }
        self._send_event(lb_const.EVENT_CREATE_HEALTH_MONITOR_V2,
                         arg_dict, serialize=True,
                         binding_key=healthmonitor[lb_const.POOL][
                             'loadbalancer_id'],
                         key=healthmonitor['id'])

    def update_healthmonitor(self, context, old_healthmonitor, healthmonitor):
        """Enqueues event for worker to process update health monitor request.

        :param context: RPC context
        :param old_health_monitor: health_monitor resource to be updated
        :param health_monitor: new health_monitor resource
        :param pool_id: pool_id to which health monitor is associated

        Returns: None

        """
        old_val, new_val = self.get_diff_of_dict(
            old_healthmonitor, healthmonitor)
        LOG.info("Received request 'Update Pool Health Monitor' for "
                 "Health monitor:%(hm)s with new Param:%(new_val)s and "
                 "old Param:%(old_val)s",
                 {'hm': healthmonitor['id'],
                  'old_val': old_val,
                  'new_val': new_val})
        arg_dict = {'context': context,
                    lb_const.OLD_HEALTHMONITOR: old_healthmonitor,
                    lb_const.HEALTHMONITOR: healthmonitor
                    }
        self._send_event(lb_const.EVENT_UPDATE_HEALTH_MONITOR_V2,
                         arg_dict, serialize=True,
                         binding_key=healthmonitor[lb_const.POOL][
                             'loadbalancer_id'],
                         key=healthmonitor['id'])

    def delete_healthmonitor(self, context, healthmonitor):
        """Enqueues event for worker to process delete health monitor request.

        :param context: RPC context
        :param health_monitor: health_monitor resource to be deleted
        :param pool_id: pool_id to which health monitor is associated

        Returns: None

        """
        LOG.info("Received request 'Delete Pool Health Monitor' for "
                 "Health monitor:%(hm)s",
                 {'hm': healthmonitor['id']})
        arg_dict = {'context': context,
                    lb_const.HEALTHMONITOR: healthmonitor
                    }
        self._send_event(lb_const.EVENT_DELETE_HEALTH_MONITOR_V2,
                         arg_dict, serialize=True,
                         binding_key=healthmonitor[lb_const.POOL][
                             'loadbalancer_id'],
                         key=healthmonitor['id'])

    def agent_updated(self, context, payload):
        """Enqueues event for worker to process agent updated request.

        :param context: RPC context
        :param payload: payload

        Returns: None

        """
        LOG.info("Received request 'Agent Updated' ")
        arg_dict = {'context': context,
                    'payload': payload}
        self._send_event(lb_const.EVENT_AGENT_UPDATED_V2, arg_dict)


"""Implements event handlers and their helper methods.

Object of this class is registered with the event class of core service
controller. Based on the event key, handle_event method of this class is
invoked by core service controller.

"""


class LBaaSV2EventHandler(agent_base.AgentBaseEventHandler,
                          nfp_api.NfpEventHandler):
    instance_mapping = {}

    def __init__(self, sc, drivers, rpcmgr):
        self.sc = sc
        self.drivers = drivers
        self.rpcmgr = rpcmgr
        self.plugin_rpc = LBaaSV2RpcSender(sc)

    def _get_driver(self, driver_name):
        """Retrieves service driver object based on service type input.

        Currently, service drivers are identified with service type. Support
        for single driver per service type is provided. When multi-vendor
        support is going to be provided, the driver should be selected based
        on both service type and vendor name.

        :param service_type: Service type - loadbalancer

        Returns: Service driver instance

        """
        driver = lb_const.SERVICE_TYPE + driver_name
        return self.drivers[driver]

    def _root_loadbalancer_id(self, obj_type, obj_dict):
        """Returns the loadbalancer id this instance is attached to."""

        try:
            # For Mitaka
            if obj_type == lb_const.LOADBALANCER:
                lb = obj_dict['id']
            elif obj_type == lb_const.LISTENER:
                lb = obj_dict[lb_const.LOADBALANCER]['id']
            elif obj_type == lb_const.L7POLICY:
                lb = obj_dict[lb_const.LISTENER][lb_const.LOADBALANCER]['id']
            elif obj_type == lb_const.L7RULE:
                lb = obj_dict['policy'][lb_const.LISTENER][
                    lb_const.LOADBALANCER]['id']
            elif obj_type == lb_const.POOL:
                lb = obj_dict[lb_const.LOADBALANCER]['id']
            elif obj_type == lb_const.SNI:
                lb = obj_dict[lb_const.LISTENER][lb_const.LOADBALANCER]['id']
            else:
                # Pool Member or Health Monitor
                lb = obj_dict[lb_const.POOL][lb_const.LOADBALANCER]['id']
            # For Liberty
            # if obj_type == lb_const.LOADBALANCER:
            #     lb = obj_dict['id']
            # elif obj_type == lb_const.LISTENER:
            #     lb = obj_dict[lb_const.LOADBALANCER]['id']
            # elif obj_type == lb_const.POOL:
            #     lb = obj_dict[lb_const.LISTENER][lb_const.LOADBALANCER]['id']
            # elif obj_type == lb_const.SNI:
            #     lb = obj_dict[lb_const.LISTENER][lb_const.LOADBALANCER]['id']
            # else:
            #     # Pool Member or Health Monitor
            #     lb = obj_dict[lb_const.POOL][lb_const.LISTENER][
            #         lb_const.LOADBALANCER]['id']
        except Exception:
            raise exceptions.IncompleteData(
                'Root loadbalancer id was not found')
        else:
            return lb

    def handle_event(self, ev):
        """Processes the generated events in worker context.

        Processes the following events.
        - create loadbalancer
        - update loadbalancer
        - delete loadbalancer
        - create listener
        - update listener
        - delete listener
        - create pool
        - update pool
        - delete pool
        - create member
        - update member
        - delete member
        - create health monitor
        - update health monitor
        - delete health monitor
        - agent updated
        Enqueues responses into notification queue.

        Returns: None

        """
        msg = ("Handling event '%s' " % (ev.id))
        LOG.info(msg)
        try:
            msg = ("Worker process with ID: %s starting "
                   "to handle task: %s of topic: %s. "
                   % (os.getpid(), ev.id, lb_const.LBAAS_AGENT_RPC_TOPIC))
            LOG.debug(msg)

            method = getattr(self, "_%s" % (ev.id.lower()))
            method(ev)
        except Exception as err:
            msg = ("Failed to perform the operation: %s. %s"
                   % (ev.id, str(err).capitalize()))
            LOG.error(msg)
        finally:
            if ev.id == lb_const.EVENT_COLLECT_STATS_V2:
                """Do not say event done for collect stats as it is
                   to be executed forever
                """
                pass
            else:
                msg = ("Calling event done for event '%s' " % (ev.id))
                LOG.info(msg)
                self.sc.event_complete(ev)

    def _handle_event_loadbalancer(self, ev, operation):
        data = ev.data
        context = data['context']
        loadbalancer = data[lb_const.LOADBALANCER]
        root_lb_id = self._root_loadbalancer_id(
            lb_const.LOADBALANCER, loadbalancer)
        agent_info = ev.data['context'].get('agent_info')
        service_vendor = agent_info['service_vendor']

        try:
            if operation == lb_const.CREATE:
                driver_name = data['driver_name']
                driver_id = driver_name + service_vendor
                if (driver_id) not in self.drivers.keys():
                    msg = ('No device driver on agent: %s.' % (driver_name))
                    LOG.error(msg)
                    self.plugin_rpc.update_status(
                        lb_const.LOADBALANCER, loadbalancer['id'], root_lb_id,
                        lb_const.ERROR, lb_const.OFFLINE, agent_info,
                        None)
                    return
                driver = self.drivers[driver_id]
                driver.load_balancer.create(context, loadbalancer)
                LBaaSV2EventHandler.instance_mapping[loadbalancer['id']] \
                    = driver_name
            elif operation == lb_const.UPDATE:
                old_loadbalancer = data[lb_const.OLD_LOADBALANCER]
                driver = self._get_driver(service_vendor)
                driver.load_balancer.update(context,
                                            old_loadbalancer, loadbalancer)
            elif operation == lb_const.DELETE:
                driver = self._get_driver(service_vendor)
                driver.load_balancer.delete(context, loadbalancer)
                del LBaaSV2EventHandler.instance_mapping[loadbalancer['id']]
                return  # Don't update object status for delete operation
        except Exception:
            if operation == lb_const.DELETE:
                msg = (
                    "Failed to delete loadbalancer %s" % (loadbalancer['id']))
                LOG.warning(msg)
                del LBaaSV2EventHandler.instance_mapping[loadbalancer['id']]
            else:
                self.plugin_rpc.update_status(
                    lb_const.LOADBALANCER, loadbalancer['id'], root_lb_id,
                    lb_const.ERROR, lb_const.OFFLINE,
                    agent_info, None)
        else:
            self.plugin_rpc.update_status(
                lb_const.LOADBALANCER, loadbalancer['id'], root_lb_id,
                lb_const.ACTIVE, lb_const.ONLINE,
                agent_info, None)

    def _create_loadbalancer_v2(self, ev):
        self._handle_event_loadbalancer(ev, lb_const.CREATE)

    def _update_loadbalancer_v2(self, ev):
        self._handle_event_loadbalancer(ev, lb_const.UPDATE)

    def _delete_loadbalancer_v2(self, ev):
        self._handle_event_loadbalancer(ev, lb_const.DELETE)

    def _handle_event_listener(self, ev, operation):
        data = ev.data
        context = data['context']
        listener = data[lb_const.LISTENER]
        root_lb_id = self._root_loadbalancer_id(lb_const.LISTENER, listener)
        agent_info = ev.data['context'].get('agent_info')
        service_vendor = agent_info['service_vendor']
        driver = self._get_driver(service_vendor)

        try:
            if operation == lb_const.CREATE:
                driver.listener.create(context, listener)
            elif operation == lb_const.UPDATE:
                old_listener = data[lb_const.OLD_LISTENER]
                driver.listener.update(context, old_listener, listener)
            elif operation == lb_const.DELETE:
                driver.listener.delete(context, listener)
                return  # Don't update object status for delete operation
        except Exception:
            if operation == lb_const.DELETE:
                msg = ("Failed to delete listener %s" % (listener['id']))
                LOG.warning(msg)
            else:
                self.plugin_rpc.update_status(
                    lb_const.LISTENER, listener['id'], root_lb_id,
                    lb_const.ERROR, lb_const.OFFLINE,
                    agent_info, None)
        else:
            self.plugin_rpc.update_status(
                lb_const.LISTENER, listener['id'], root_lb_id,
                lb_const.ACTIVE, lb_const.ONLINE,
                agent_info, None)

    def _create_listener_v2(self, ev):
        self._handle_event_listener(ev, lb_const.CREATE)

    def _update_listener_v2(self, ev):
        self._handle_event_listener(ev, lb_const.UPDATE)

    def _delete_listener_v2(self, ev):
        self._handle_event_listener(ev, lb_const.DELETE)

    def _handle_event_pool(self, ev, operation):
        data = ev.data
        context = data['context']
        pool = data[lb_const.POOL]
        root_lb_id = self._root_loadbalancer_id(lb_const.POOL, pool)
        agent_info = ev.data['context'].get('agent_info')
        service_vendor = agent_info['service_vendor']
        driver = self._get_driver(service_vendor)

        try:
            if operation == lb_const.CREATE:
                driver.pool.create(context, pool)
            elif operation == lb_const.UPDATE:
                old_pool = data[lb_const.OLD_POOL]
                driver.pool.update(context, old_pool, pool)
            elif operation == lb_const.DELETE:
                driver.pool.delete(context, pool)
                return  # Don't update object status for delete operation
        except Exception:
            if operation == lb_const.DELETE:
                msg = "Failed to delete pool %s" % (pool['id'])
                LOG.warning(msg)
            else:
                self.plugin_rpc.update_status(
                    lb_const.POOL, pool['id'], root_lb_id,
                    lb_const.ERROR, lb_const.OFFLINE,
                    agent_info, None)
        else:
            self.plugin_rpc.update_status(
                lb_const.POOL, pool['id'], root_lb_id,
                lb_const.ACTIVE, lb_const.ONLINE,
                agent_info, None)

    def _create_pool_v2(self, ev):
        self._handle_event_pool(ev, lb_const.CREATE)

    def _update_pool_v2(self, ev):
        self._handle_event_pool(ev, lb_const.UPDATE)

    def _delete_pool_v2(self, ev):
        self._handle_event_pool(ev, lb_const.DELETE)

    def _handle_event_member(self, ev, operation):
        data = ev.data
        context = data['context']
        member = data[lb_const.MEMBER]
        root_lb_id = self._root_loadbalancer_id(lb_const.MEMBER, member)
        agent_info = ev.data['context'].get('agent_info')
        service_vendor = agent_info['service_vendor']
        driver = self._get_driver(service_vendor)  # member['pool_id'])
        try:
            if operation == lb_const.CREATE:
                driver.member.create(context, member)
            elif operation == lb_const.UPDATE:
                old_member = data[lb_const.OLD_MEMBER]
                driver.member.update(context, old_member, member)
            elif operation == lb_const.DELETE:
                driver.member.delete(context, member)
                return  # Don't update object status for delete operation
        except Exception:
            if operation == lb_const.DELETE:
                msg = ("Failed to delete member %s" % (member['id']))
                LOG.warning(msg)
            else:
                self.plugin_rpc.update_status(
                    lb_const.MEMBER, member['id'], root_lb_id,
                    lb_const.ERROR, lb_const.OFFLINE,
                    agent_info, None)
        else:
            self.plugin_rpc.update_status(
                lb_const.MEMBER, member['id'], root_lb_id,
                lb_const.ACTIVE, lb_const.ONLINE,
                agent_info, None)

    def _create_member_v2(self, ev):
        self._handle_event_member(ev, lb_const.CREATE)

    def _update_member_v2(self, ev):
        self._handle_event_member(ev, lb_const.UPDATE)

    def _delete_member_v2(self, ev):
        self._handle_event_member(ev, lb_const.DELETE)

    def _handle_event_health_monitor(self, ev, operation):
        data = ev.data
        context = data['context']
        healthmonitor = data[lb_const.HEALTHMONITOR]
        root_lb_id = self._root_loadbalancer_id(
            lb_const.HEALTHMONITOR, healthmonitor)
        agent_info = context.get('agent_info')
        service_vendor = agent_info['service_vendor']
        driver = self._get_driver(service_vendor)  # (pool_id)

        pool_id = healthmonitor[lb_const.POOL]['id']
        assoc_id = {'pool_id': pool_id,
                    'monitor_id': healthmonitor['id']}
        try:
            if operation == lb_const.CREATE:
                driver.health_monitor.create(context, healthmonitor)
            elif operation == lb_const.UPDATE:
                old_healthmonitor = data[lb_const.OLD_HEALTHMONITOR]
                driver.health_monitor.update(context, old_healthmonitor,
                                             healthmonitor)
            elif operation == lb_const.DELETE:
                driver.health_monitor.delete(context, healthmonitor)
                return  # Don't update object status for delete operation
        except Exception:
            if operation == lb_const.DELETE:
                msg = ("Failed to delete pool health monitor."
                       " assoc_id: %s" % (assoc_id))
                LOG.warning(msg)
            else:
                self.plugin_rpc.update_status(
                    lb_const.HEALTHMONITOR, healthmonitor['id'], root_lb_id,
                    lb_const.ERROR, lb_const.OFFLINE,
                    agent_info, None)
        else:
            self.plugin_rpc.update_status(
                lb_const.HEALTHMONITOR, healthmonitor['id'], root_lb_id,
                lb_const.ACTIVE, lb_const.ONLINE,
                agent_info, None)

    def _create_health_monitor_v2(self, ev):
        self._handle_event_health_monitor(ev, lb_const.CREATE)

    def _update_health_monitor_v2(self, ev):
        self._handle_event_health_monitor(ev, lb_const.UPDATE)

    def _delete_health_monitor_v2(self, ev):
        self._handle_event_health_monitor(ev, lb_const.DELETE)

    def _agent_updated(self, ev):
        """ REVISIT(pritam): Support """
        return None

    def _collect_stats(self, ev):
        self.sc.poll_event(ev)

    @nfp_api.poll_event_desc(event=lb_const.EVENT_COLLECT_STATS_V2,
                             spacing=60)
    def collect_stats_v2(self, ev):
        for pool_id, driver_name in \
                LBaaSV2EventHandler.instance_mapping.items():
            driver_id = lb_const.SERVICE_TYPE + driver_name
            driver = self.drivers[driver_id]
            try:
                stats = driver.get_stats(pool_id)
                if stats:
                    self.plugin_rpc.update_pool_stats(pool_id, stats,
                                                      self.context)
            except Exception:
                msg = ("Error updating statistics on pool %s" % (pool_id))
                LOG.error(msg)


def events_init(sc, drivers, rpcmgr):
    """Registers events with core service controller.

    All the events will come to handle_event method of class instance
    registered in 'handler' field.

    :param drivers: Driver instances registered with the service agent
    :param rpcmgr: Instance to receive all the RPC messages from configurator
    module.

    Returns: None

    """
    ev_ids = [lb_const.EVENT_CREATE_LOADBALANCER_V2,
              lb_const.EVENT_UPDATE_LOADBALANCER_V2,
              lb_const.EVENT_DELETE_LOADBALANCER_V2,

              lb_const.EVENT_CREATE_LISTENER_V2,
              lb_const.EVENT_UPDATE_LISTENER_V2,
              lb_const.EVENT_DELETE_LISTENER_V2,

              lb_const.EVENT_CREATE_POOL_V2, lb_const.EVENT_UPDATE_POOL_V2,
              lb_const.EVENT_DELETE_POOL_V2,

              lb_const.EVENT_CREATE_MEMBER_V2,
              lb_const.EVENT_UPDATE_MEMBER_V2,
              lb_const.EVENT_DELETE_MEMBER_V2,

              lb_const.EVENT_CREATE_HEALTH_MONITOR_V2,
              lb_const.EVENT_UPDATE_HEALTH_MONITOR_V2,
              lb_const.EVENT_DELETE_HEALTH_MONITOR_V2,

              lb_const.EVENT_AGENT_UPDATED_V2,
              lb_const.EVENT_COLLECT_STATS_V2
              ]

    evs = []
    for ev_id in ev_ids:
        ev = nfp_event.Event(id=ev_id, handler=LBaaSV2EventHandler(
            sc, drivers, rpcmgr))
        evs.append(ev)
    sc.register_events(evs)


def load_drivers(sc, conf):
    """Imports all the driver files.

    Returns: Dictionary of driver objects with a specified service type and/or
    vendor name

    """
    cutils = utils.ConfiguratorUtils(conf)
    drivers = cutils.load_drivers(lb_const.SERVICE_TYPE)

    plugin_rpc = LBaaSV2RpcSender(sc)

    for service_type, dobj in six.iteritems(drivers):
        '''LB Driver constructor needs plugin_rpc as a param'''
        instantiated_dobj = dobj(plugin_rpc=plugin_rpc, conf=conf)
        drivers[service_type] = instantiated_dobj

    return drivers


def register_service_agent(cm, sc, conf, rpcmgr):
    """Registers Loadbalaner V2 service agent with configurator module.

    :param cm: Instance of configurator module
    :param sc: Instance of core service controller
    :param conf: Instance of oslo configuration
    :param rpcmgr: Instance containing RPC methods which are invoked by
    configurator module on corresponding RPC message arrival

    """

    service_type = lb_const.SERVICE_TYPE
    cm.register_service_agent(service_type, rpcmgr)


def init_agent(cm, sc, conf):
    """Initializes Loadbalaner V2 agent.

    :param cm: Instance of configuration module
    :param sc: Instance of core service controller
    :param conf: Instance of oslo configuration

    """

    try:
        drivers = load_drivers(sc, conf)
    except Exception as err:
        msg = ("Loadbalaner V2 agent failed to load service drivers. %s"
               % (str(err).capitalize()))
        LOG.error(msg)
        raise err
    else:
        msg = ("Loadbalaner V2 agent loaded service"
               " drivers successfully.")
        LOG.debug(msg)

    rpcmgr = LBaaSv2RpcManager(sc, conf)

    try:
        events_init(sc, drivers, rpcmgr)
    except Exception as err:
        msg = ("Loadbalaner V2 agent failed to initialize events. %s"
               % (str(err).capitalize()))
        LOG.error(msg)
        raise err
    else:
        msg = ("Loadbalaner V2 agent initialized"
               " events successfully.")
        LOG.debug(msg)

    try:
        register_service_agent(cm, sc, conf, rpcmgr)
    except Exception as err:
        msg = ("Failed to register Loadbalaner V2 agent with"
               " configurator module. %s" % (str(err).capitalize()))
        LOG.error(msg)
        raise err
    else:
        msg = ("Loadbalaner V2 agent registered with configuration"
               " module successfully.")
        LOG.debug(msg)


def _start_collect_stats(sc):
    """Enqueues poll event for worker to collect pool stats periodically.
       Agent keeps map of pool_id:driver. As part of handling this event,
       stats for pool_id is requested from agent inside service vm
    """

    arg_dict = {}
    ev = sc.new_event(id=lb_const.EVENT_COLLECT_STATS_V2, data=arg_dict)
    sc.post_event(ev)


def init_agent_complete(cm, sc, conf):
    # _start_collect_stats(sc)
    msg = ("Initialization of loadbalancer agent v2 completed.")
    LOG.info(msg)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import uuid as pyuuid

from keystoneclient.v2_0 import client as identity_client
import mock
from oslo_config import cfg

from gbpservice.neutron.tests.unit.nfp.orchestrator.db import test_nfp_db
from gbpservice.nfp.common import constants as nfp_constants
from gbpservice.nfp.common import exceptions as nfp_exc
from gbpservice.nfp.core import context as nfp_context
from gbpservice.nfp.core import controller  # noqa
from gbpservice.nfp.core.event import Event as NFP_EVENT
from gbpservice.nfp.lib import nfp_context_manager
from gbpservice.nfp.lib import transport
from gbpservice.nfp.orchestrator.modules import (
    service_orchestrator as nso)
from gbpservice.nfp.orchestrator.openstack import openstack_driver


nfp_context_manager.sql_lock_support = False


def Event(**kwargs):
    data = kwargs.get('data')
    key = pyuuid.uuid4()
    return NFP_EVENT(key=key,
                     id='',
                     data=data)


class NSOModuleTestCase(test_nfp_db.NFPDBTestCase):

    def setUp(self):
        super(NSOModuleTestCase, self).setUp()

    @mock.patch.object(nso, 'events_init')
    @mock.patch.object(nso, 'rpc_init')
    def test_module_init(self, mock_rpc_init, mock_events_init):
        controller = mock.Mock()
        with mock.patch.object(identity_client, "Client"):
            nso.nfp_module_init(controller, cfg.CONF)
            mock_events_init.assert_called_once_with(
                controller, cfg.CONF, mock.ANY)
            call_args, call_kwargs = mock_events_init.call_args
            self.assertIsInstance(call_args[2],
                                  nso.ServiceOrchestrator)
            mock_rpc_init.assert_called_once_with(controller, cfg.CONF)

    def test_rpc_init(self):
        controller = mock.Mock()
        nso.rpc_init(controller, cfg.CONF)
        controller.register_rpc_agents.assert_called_once_with(mock.ANY)
        call_args, call_kwargs = controller.register_rpc_agents.call_args
        self.assertEqual(2, len(call_args[0]))
        self.assertIsInstance(call_args[0][0], nso.RpcAgent)

    def test_events_init(self):
        controller = mock.Mock()
        with mock.patch.object(identity_client, "Client"):
            nso.events_init(
                controller, cfg.CONF,
                nso.ServiceOrchestrator(controller, cfg.CONF))
            controller.register_events.assert_called_once_with(mock.ANY)


class NSORpcHandlerTestCase(NSOModuleTestCase):

    def setUp(self):
        super(NSORpcHandlerTestCase, self).setUp()
        self.controller = mock.Mock()
        self.rpc_handler = nso.RpcHandler(cfg.CONF, self.controller)

    @mock.patch.object(nso.ServiceOrchestrator,
                       "create_network_function")
    def test_rpc_create_network_function(self, mock_create_network_function):
        with mock.patch.object(identity_client, "Client"):
            self.rpc_handler.create_network_function(
                "context", {'resource_owner_context':
                            {'tenant_id': 'tenant_id'}})
            mock_create_network_function.assert_called_once_with(
                "context", mock.ANY)

    @mock.patch.object(nso.ServiceOrchestrator,
                       "get_network_function")
    def test_rpc_get_network_function(self, mock_get_network_function):
        with mock.patch.object(identity_client, "Client"):
            self.rpc_handler.get_network_function(
                "context", "network_function_id")
            mock_get_network_function.assert_called_once_with(
                "context", "network_function_id")

    @mock.patch.object(nso.ServiceOrchestrator,
                       "get_network_functions")
    def test_rpc_get_network_functions(self, mock_get_network_functions):
        with mock.patch.object(identity_client, "Client"):
            filters = {'id': ['myuuid']}
            self.rpc_handler.get_network_functions("context", filters=filters)
            mock_get_network_functions.assert_called_once_with(
                "context", filters)

    @mock.patch.object(nso.ServiceOrchestrator,
                       "delete_network_function")
    def test_rpc_delete_network_function(self, mock_delete_network_function):
        with mock.patch.object(identity_client, "Client"):
            self.rpc_handler.delete_network_function(
                "context", "network_function_id", mock.ANY)
            mock_delete_network_function.assert_called_once_with(
                "context", "network_function_id", mock.ANY)

    @mock.patch.object(nso.ServiceOrchestrator,
                       "update_network_function")
    def test_rpc_update_network_function(self, mock_update_network_function):
        with mock.patch.object(identity_client, "Client"):
            self.rpc_handler.update_network_function(
                "context", "network_function_id", "updated_network_function")
            mock_update_network_function.assert_called_once_with(
                "context", "network_function_id", "updated_network_function")

    @mock.patch.object(nso.ServiceOrchestrator,
                       "handle_policy_target_added")
    def test_rpc_policy_target_added_notification(
            self, mock_handle_policy_target_added):
        with mock.patch.object(identity_client, "Client"):
            self.rpc_handler.policy_target_added_notification(
                "context", "network_function_id", "policy_target")
            mock_handle_policy_target_added.assert_called_once_with(
                "context", "network_function_id", "policy_target")

    @mock.patch.object(nso.ServiceOrchestrator,
                       "handle_policy_target_removed")
    def test_rpc_policy_target_removed_notification(
            self, mock_handle_policy_target_removed):
        with mock.patch.object(identity_client, "Client"):
            self.rpc_handler.policy_target_removed_notification(
                "context", "network_function_id", "policy_target")
            mock_handle_policy_target_removed.assert_called_once_with(
                "context", "network_function_id", "policy_target")

    @mock.patch.object(
        nso.ServiceOrchestrator, "handle_consumer_ptg_added")
    def test_rpc_consumer_ptg_added_notification(
            self, mock_handle_consumer_ptg_added):
        with mock.patch.object(identity_client, "Client"):
            self.rpc_handler.consumer_ptg_added_notification(
                "context", "network_function_id", "policy_target_group")
            mock_handle_consumer_ptg_added.assert_called_once_with(
                "context", "network_function_id", "policy_target_group")

    @mock.patch.object(
        nso.ServiceOrchestrator, "handle_consumer_ptg_removed")
    def test_rpc_consumer_ptg_removed_notification(
            self, mock_handle_consumer_ptg_removed):
        with mock.patch.object(identity_client, "Client"):
            self.rpc_handler.consumer_ptg_removed_notification(
                "context", "network_function_id", "policy_target_group")
            mock_handle_consumer_ptg_removed.assert_called_once_with(
                "context", "network_function_id", "policy_target_group")


class ServiceOrchestratorTestCase(NSOModuleTestCase):

    def setUp(self):
        super(ServiceOrchestratorTestCase, self).setUp()
        self.controller = mock.Mock()
        self.context = mock.Mock()
        cfg.CONF.set_override("auth_version", "v1",
                              group="nfp_keystone_authtoken")
        with mock.patch.object(identity_client, "Client"):
            self.service_orchestrator = nso.ServiceOrchestrator(
                self.controller,
                cfg.CONF)
            self.service_orchestrator.config_driver.\
                parse_template_config_string = (mock.Mock(return_value=(
                    'heat_config',
                    'template')))

    @mock.patch.object(
        openstack_driver.KeystoneClient, "get_admin_tenant_id")
    @mock.patch.object(
        openstack_driver.KeystoneClient, "get_admin_token")
    @mock.patch.object(
        openstack_driver.GBPClient, "get_service_profile")
    @mock.patch.object(
        nso.ServiceOrchestrator, "_create_event")
    @mock.patch.object(
        nso.NSOConfiguratorRpcApi, "create_network_function_user_config")
    def test_create_network_function(self, mock_rpc, mock_create_event,
                                     mock_get_service_profile,
                                     mock_get_admin_token,
                                     mock_get_admin_tenant_id):
        (self.service_orchestrator.db_handler.
            update_node_instance_network_function_map) = mock.MagicMock(
                return_value=None)
        network_function_info = {
            'tenant_id': 'tenant_id',
            'service_chain_id': 'sc_instance_id',
            'service_id': 'sc_node_id',
            'service_profile_id': 'service_profile_id',
            'management_ptg_id': 'mgmt_ptg_id',
            'service_cfg.CONF': '',
            'port_info': {
                'id': 'provider_port_id',
                'port_model': nfp_constants.GBP_PORT,
                'port_classification': nfp_constants.PROVIDER
            },
            'network_function_mode': nfp_constants.GBP_MODE,
            'provider': None,
            'consumer': None,
            'resource_owner_context': {'admin_token': str(pyuuid.uuid4()),
                                       'admin_tenant_id': str(pyuuid.uuid4())},
            'service_chain_instance': {'id': str(pyuuid.uuid4()),
                                       'name': str(pyuuid.uuid4())},
            'service_chain_node': {'id': str(pyuuid.uuid4()),
                                   'name': str(pyuuid.uuid4())},
            'service_profile': {'id': str(pyuuid.uuid4()),
                                'service_flavor': None,
                                'service_type': 'xyz'},
            'service_config': None,
            'network_function_mode': 'gbp'
        }
        transport.parse_service_flavor_string = mock.MagicMock(
            return_value={'device_type': 'None',
                          'service_vendor': 'vyos'})
        self.service_orchestrator.config_driver.\
            parse_template_config_string = mock.MagicMock(
                return_value=('heat_config', '{}'))
        network_function = self.service_orchestrator.create_network_function(
            self.context, network_function_info)
        self.assertIsNotNone(network_function)
        db_network_function = self.nfp_db.get_network_function(
            self.session, network_function['id'])
        service_config = db_network_function.pop('service_config')
        self.assertIsNone(service_config)
        self.assertEqual(network_function, db_network_function)

    def test_validate_create_service_input(self):
        network_function = {}
        self.assertRaises(
            nfp_exc.RequiredDataNotProvided,
            self.service_orchestrator._validate_create_service_input,
            self.context, network_function)

        network_function = {
            "tenant_id": "test",
            "service_id": "test",
            "service_chain_id": "test",
            "service_profile_id": "test",
            "network_function_mode": "test",
            'network_function_mode': nfp_constants.GBP_MODE,
            'provider': None,
            'consumer': None,
            'resource_owner_context': {'admin_token': str(pyuuid.uuid4()),
                                       'admin_tenant_id': str(pyuuid.uuid4())},
            'service_chain_instance': {'id': str(pyuuid.uuid4())},
            'service_chain_node': {'id': str(pyuuid.uuid4())},
            'service_profile': {'id': str(pyuuid.uuid4()),
                                'service_flavor': None, 'service_type': 'xyz'},
            'service_config': None,
            'network_function_mode': 'gbp',
            'management_ptg_id': None
        }
        transport.parse_service_flavor_string = mock.MagicMock(
            return_value={'device_type': 'VM',
                          'service_vendor': 'vyos'})
        return_value = (
            self.service_orchestrator._validate_create_service_input(
                self.context, network_function))
        self.assertIsNone(return_value)

    @mock.patch.object(
        openstack_driver.KeystoneClient, "get_admin_tenant_id")
    @mock.patch.object(
        openstack_driver.KeystoneClient, "get_admin_token")
    @mock.patch.object(
        openstack_driver.GBPClient, "get_service_profile")
    @mock.patch.object(
        nso.NSOConfiguratorRpcApi, "delete_network_function_user_config")
    def test_delete_network_function_without_nfi(self, mock_rpc,
                                                 mock_get_service_profile,
                                                 mock_get_admin_token,
                                                 mock_get_admin_tenant_id):
        network_function = self.create_network_function()
        nfp_context.init()
        mock_get_admin_token.return_value = 'admin_token'
        mock_get_admin_tenant_id.return_value = 'admin_tenant_id'
        transport.parse_service_flavor_string = mock.MagicMock(
            return_value={'device_type': 'VM',
                          'service_vendor': 'vyos'})
        self.service_orchestrator._create_event = mock.MagicMock(
            return_value='')

        delete_network_function_data = {
            'service_chain_instance': {},
            'provider': {},
            'consumer': {}}
        self.service_orchestrator.delete_network_function(
            self.context, network_function['id'], delete_network_function_data)
        self.assertRaises(nfp_exc.NetworkFunctionNotFound,
                          self.nfp_db.get_network_function,
                          self.session, network_function['id'])
        self.assertFalse(self.controller.event.called)
        self.assertFalse(self.controller.rpc_event.called)

    @mock.patch.object(
        openstack_driver.KeystoneClient, "get_admin_tenant_id")
    @mock.patch.object(
        nso.ServiceOrchestrator, "_create_event")
    @mock.patch.object(
        openstack_driver.KeystoneClient, "get_admin_token")
    @mock.patch.object(
        openstack_driver.GBPClient, "get_service_profile")
    @mock.patch.object(
        nso.NSOConfiguratorRpcApi, "delete_network_function_user_config")
    def test_delete_network_function_with_nfi(self, mock_rpc,
                                              mock_get_service_profile,
                                              mock_get_admin_token,
                                              mock_create_event,
                                              mock_get_admin_tenant_id):
        (self.service_orchestrator.db_handler.
            update_node_instance_network_function_map) = mock.MagicMock(
                return_value=None)
        network_function_instance = self.create_network_function_instance()
        network_function_id = network_function_instance['network_function_id']
        network_function = self.nfp_db.get_network_function(
            self.session, network_function_id)
        mock_get_admin_token.return_value = 'admin_token'
        mock_get_admin_tenant_id.return_value = 'admin_tenant_id'
        nfp_context.init()

        transport.parse_service_flavor_string = mock.MagicMock(
            return_value={'device_type': 'VM',
                          'service_vendor': 'vyos'})
        delete_network_function_data = {
            'service_chain_instance': {},
            'provider': {},
            'consumer': {}}
        self.service_orchestrator.delete_network_function(
            self.context, network_function_id, delete_network_function_data)
        network_function = self.nfp_db.get_network_function(
            self.session, network_function_id)
        self.assertEqual('PENDING_DELETE', network_function['status'])

    def test_event_create_network_function_instance(self):
        network_function_instance = self.create_network_function_instance()
        network_function = self.nfp_db.get_network_function(
            self.session,
            network_function_instance['network_function_id'])
        network_function_port_info = [
            {
                'id': 'provider_port_id',
                'port_model': nfp_constants.GBP_PORT,
                'port_classification': nfp_constants.PROVIDER
            },
            {
                'id': 'consumer_port_id',
                'port_model': nfp_constants.GBP_PORT,
                'port_classification': nfp_constants.CONSUMER
            }
        ]
        management_network_info = {
            'id': 'management_ptg_id',
            'port_model': nfp_constants.GBP_PORT
        }

        create_nfi_request = {
            'network_function': network_function,
            'network_function_port_info': network_function_port_info,
            'network_function_instance': network_function_instance,
            'management_network_info': management_network_info,
            'service_type': 'service_type',
            'service_details': {'service_vendor': 'vendor',
                                'service_type': 'xyz'},
            'service_vendor': 'vendor',
            'share_existing_device': True,
            'service_profile': None,
            'consumer': {'pt': None},
            'provider': {'pt': None}
        }
        test_event = Event(data=create_nfi_request)
        test_event.context = create_nfi_request
        test_event.context['log_context'] = nfp_context.init_log_context()
        self.service_orchestrator.create_network_function_instance(
            test_event)

    @mock.patch.object(
        nso.ServiceOrchestrator, "_create_event")
    @mock.patch.object(
        openstack_driver.KeystoneClient, "get_admin_token")
    @mock.patch.object(
        openstack_driver.GBPClient, "get_service_profile")
    @mock.patch.object(
        nso.NSOConfiguratorRpcApi, "create_network_function_user_config")
    def test_event_handle_device_active(self, mock_create_rpc,
                                        mock_service_profile,
                                        mock_admin_token,
                                        mock_create_event):
        nfd = self.create_network_function_device()
        nfi = self.create_network_function_instance(create_nfd=False)
        request_data = {
            'network_function_instance_id': nfi['id'],
            'network_function_device_id': nfd['id']
        }
        test_event = Event(data=request_data)
        self.assertIsNone(nfi['network_function_device_id'])
        with mock.patch.object(
                self.service_orchestrator.config_driver,
                "apply_config") as mock_apply_user_config:
            mock_apply_user_config.return_value = "stack_id"
            self.service_orchestrator.handle_device_active(
                test_event)
        db_nfi = self.nfp_db.get_network_function_instance(
            self.session, nfi['id'])
        db_nf = self.nfp_db.get_network_function(
            self.session, nfi['network_function_id'])
        self.assertEqual(nfd['id'], db_nfi['network_function_device_id'])
        self.assertIsNotNone(db_nf['config_policy_id'])

    def test_event_handle_device_create_failed(self):
        (self.service_orchestrator.db_handler.
            update_node_instance_network_function_map) = mock.MagicMock(
                return_value=None)
        nfd = self.create_network_function_device()
        nfi = self.create_network_function_instance(create_nfd=False)
        request_data = {
            'network_function_instance_id': nfi['id'],
            'network_function_device_id': nfd['id']
        }
        test_event = Event(data=request_data, context=request_data)
        test_event.context['log_context'] = nfp_context.init_log_context()
        self.assertIsNone(nfi['network_function_device_id'])
        self.service_orchestrator.handle_device_create_failed(
            test_event)
        db_nfi = self.nfp_db.get_network_function_instance(
            self.session, nfi['id'])
        db_nf = self.nfp_db.get_network_function(
            self.session, nfi['network_function_id'])
        self.assertEqual(nfp_constants.ERROR, db_nfi['status'])
        self.assertEqual(nfp_constants.ERROR, db_nf['status'])

    def test_event_check_for_user_config_complete(self):
        (self.service_orchestrator.db_handler.
            update_node_instance_network_function_map) = mock.MagicMock(
                return_value=None)
        network_function = self.create_network_function()
        network_function_details = (
            self.service_orchestrator.get_network_function_details(
                network_function['id']))
        with mock.patch.object(
                self.service_orchestrator.config_driver,
                "check_config_complete") as mock_is_config_complete,\
                mock.patch.object(identity_client, "Client"):
            # Verify return status IN_PROGRESS from cfg.CONF driver
            mock_is_config_complete.return_value = "IN_PROGRESS"
            request_data = {
                'service_details': {'service_vendor': 'vyos'},
                'tenant_id': network_function['tenant_id'],
                'config_policy_id': 'config_policy_id',
                'network_function_id': network_function['id'],
                'network_function_details': network_function_details,
                'network_function': {'id': network_function['id']},
                'event_desc': {'poll_desc': None, 'worker': None,
                               'flag': None, 'type': None,
                               'uuid': 'a1251c79-f661-440e-aab2-a1f401865daf:'}
            }
            test_event = Event(data=request_data)
            test_event.context = request_data
            status = self.service_orchestrator.check_for_user_config_complete(
                test_event)
            mock_is_config_complete.assert_called_once_with(
                request_data)
            self.nfp_db.get_network_function(
                self.session, network_function['id'])
            self.assertEqual(status, nso.CONTINUE_POLLING)

            # Verify return status ERROR from cfg.CONF driver
            mock_is_config_complete.reset_mock()
            mock_is_config_complete.return_value = "ERROR"
            request_data = {
                'service_details': {'service_vendor': 'vyos'},
                'tenant_id': network_function['tenant_id'],
                'config_policy_id': 'config_policy_id',
                'network_function_id': network_function['id'],
                'network_function_details': network_function_details,
                'network_function': {'id': network_function['id']},
                'event_desc': {'poll_desc': None, 'worker': None,
                               'flag': None, 'type': None,
                               'uuid': 'a1251c79-f661-440e-aab2-a1f401865daf:'}
            }
            test_event = Event(data=request_data)
            test_event.context = request_data
            status = self.service_orchestrator.check_for_user_config_complete(
                test_event)
            mock_is_config_complete.assert_called_once_with(
                request_data)
            self.nfp_db.get_network_function(
                self.session, network_function['id'])
            self.assertEqual(status, nso.STOP_POLLING)

            # Verify return status COMPLETED from cfg.CONF driver
            self.controller.poll_event_done.reset_mock()
            mock_is_config_complete.reset_mock()
            mock_is_config_complete.return_value = "COMPLETED"
            request_data = {
                'service_details': {'service_vendor': 'vyos'},
                'tenant_id': network_function['tenant_id'],
                'config_policy_id': 'config_policy_id',
                'network_function_id': network_function['id'],
                'network_function_details': network_function_details,
                'network_function': {'id': network_function['id']},
                'event_desc': {'poll_desc': None, 'worker': None,
                               'flag': None, 'type': None,
                               'uuid': 'a1251c79-f661-440e-aab2-a1f401865daf:'}
            }
            test_event = Event(data=request_data)
            test_event.context = request_data
            status = self.service_orchestrator.check_for_user_config_complete(
                test_event)
            mock_is_config_complete.assert_called_once_with(
                request_data)
            self.nfp_db.get_network_function(
                self.session, network_function['id'])
            self.assertEqual(status, nso.STOP_POLLING)

    def test_event_handle_user_config_applied(self):
        (self.service_orchestrator.db_handler.
            update_node_instance_network_function_map) = mock.MagicMock(
                return_value=None)
        network_function = self.create_network_function()
        request_data = {
            'config_policy_id': 'config_policy_id',
            'network_function_id': network_function['id']
        }
        test_event = Event(data=request_data)
        self.service_orchestrator.handle_user_config_applied(test_event)
        db_nf = self.nfp_db.get_network_function(
            self.session, network_function['id'])
        self.assertEqual('ACTIVE', db_nf['status'])

    def test_event_handle_user_config_failed(self):
        (self.service_orchestrator.db_handler.
            update_node_instance_network_function_map) = mock.MagicMock(
                return_value=None)
        network_function = self.create_network_function()
        request_data = {
            'config_policy_id': 'config_policy_id',
            'network_function_id': network_function['id']
        }
        test_event = Event(data=request_data, context=request_data)
        test_event.context['log_context'] = nfp_context.init_log_context()
        self.service_orchestrator.handle_user_config_failed(test_event)
        db_nf = self.nfp_db.get_network_function(
            self.session, network_function['id'])
        self.assertEqual('ERROR', db_nf['status'])

    @mock.patch.object(
        nso.ServiceOrchestrator, "_get_service_type")
    @mock.patch.object(
        nso.ServiceOrchestrator, "_create_event")
    def test_event_check_for_user_config_deleted(self, mock_create_event,
                                                 mock_service_type):
        (self.service_orchestrator.db_handler.
            update_node_instance_network_function_map) = mock.MagicMock(
                return_value=None)
        network_function = self.create_network_function()
        with mock.patch.object(
                self.service_orchestrator.config_driver,
                "is_config_delete_complete") as mock_is_config_delete_complete:
            # Verify return status IN_PROGRESS from cfg.CONF driver
            mock_is_config_delete_complete.return_value = "IN_PROGRESS"
            request_data = {
                'tenant_id': network_function['tenant_id'],
                'config_policy_id': 'config_policy_id',
                'network_function_id': network_function['id']}
            test_event = Event(data=request_data)
            test_event.context = request_data
            mock_service_type.return_value = 'firewall'
            status = self.service_orchestrator.check_for_user_config_deleted(
                test_event)
            mock_is_config_delete_complete.assert_called_once_with(
                request_data['config_policy_id'],
                network_function['tenant_id'],
                network_function)
            db_nf = self.nfp_db.get_network_function(
                self.session, network_function['id'])
            self.assertEqual(network_function['status'], db_nf['status'])
            self.assertEqual(network_function['config_policy_id'],
                             db_nf['config_policy_id'])
            self.assertEqual(status, nso.CONTINUE_POLLING)

            # Verify return status ERROR from cfg.CONF driver
            mock_is_config_delete_complete.reset_mock()
            mock_is_config_delete_complete.return_value = "ERROR"
            request_data = {
                'tenant_id': network_function['tenant_id'],
                'config_policy_id': 'config_policy_id',
                'network_function_id': network_function['id']}
            test_event = Event(data=request_data)
            test_event.context = request_data
            status = self.service_orchestrator.check_for_user_config_deleted(
                test_event)
            mock_is_config_delete_complete.assert_called_once_with(
                request_data['config_policy_id'],
                network_function['tenant_id'],
                network_function)
            mock_create_event.assert_called_once_with(
                'USER_CONFIG_DELETE_FAILED', event_data=request_data,
                is_internal_event=True)
            self.assertEqual(status, nso.STOP_POLLING)

            # Verify return status COMPLETED from cfg.CONF driver
            self.controller.poll_event_done.reset_mock()
            mock_is_config_delete_complete.reset_mock()
            mock_create_event.reset_mock()
            mock_is_config_delete_complete.return_value = "COMPLETED"
            request_data = {
                'tenant_id': network_function['tenant_id'],
                'config_policy_id': 'config_policy_id',
                'network_function_id': network_function['id'],
                'action': 'update'}
            test_event = Event(data=request_data)
            test_event.context = request_data
            test_event.context['log_context'] = nfp_context.init_log_context()
            status = self.service_orchestrator.check_for_user_config_deleted(
                test_event)
            mock_is_config_delete_complete.assert_called_once_with(
                request_data['config_policy_id'],
                network_function['tenant_id'],
                network_function)
            db_nf = self.nfp_db.get_network_function(
                self.session, network_function['id'])
            self.assertEqual('config_policy_id', db_nf['config_policy_id'])
            self.assertEqual(status, nso.STOP_POLLING)

    def test_event_handle_user_config_delete_failed(self):
        (self.service_orchestrator.db_handler.
            update_node_instance_network_function_map) = mock.MagicMock(
                return_value=None)
        network_function = self.create_network_function()
        request_data = {
            'network_function_id': network_function['id']
        }
        test_event = Event(data=request_data)
        self.service_orchestrator.handle_user_config_delete_failed(test_event)
        db_nf = self.nfp_db.get_network_function(
            self.session, network_function['id'])
        self.assertEqual('ERROR', db_nf['status'])

    @mock.patch.object(
        openstack_driver.KeystoneClient, "get_admin_tenant_id")
    @mock.patch.object(
        nso.ServiceOrchestrator, "_create_event")
    @mock.patch.object(
        openstack_driver.KeystoneClient, "get_admin_token")
    @mock.patch.object(
        openstack_driver.GBPClient, "get_service_profile")
    @mock.patch.object(
        nso.NSOConfiguratorRpcApi, "delete_network_function_user_config")
    def test_delete_network_function(self, mock_rpc, mock_get_service_profile,
                                     mock_get_admin_token,
                                     mock_create_event,
                                     mock_get_admin_tenant_id):
        nfi = self.create_network_function_instance()
        network_function = self.nfp_db.get_network_function(
            self.session, nfi['network_function_id'])
        transport.parse_service_flavor_string = mock.MagicMock(
            return_value={'device_type': 'VM',
                          'service_vendor': 'vyos'})
        self.assertEqual([nfi['id']],
                         network_function['network_function_instances'])
        mock_get_admin_token.return_value = 'admin_token'
        mock_get_admin_tenant_id.return_value = 'admin_tenant_id'
        nfp_context.init()
        (self.service_orchestrator.db_handler.
            update_node_instance_network_function_map) = mock.MagicMock(
                return_value=None)
        delete_network_function_data = {
            'service_chain_instance': {},
            'provider': {},
            'consumer': {}}
        self.service_orchestrator.delete_network_function(
            self.context, network_function['id'], delete_network_function_data)
        db_nf = self.nfp_db.get_network_function(
            self.session, network_function['id'])
        self.assertEqual('PENDING_DELETE', db_nf['status'])

    @mock.patch.object(
        nso.ServiceOrchestrator, "_create_event")
    def test_event_delete_network_function_instance(self, mock_create_event):
        nfi = self.create_network_function_instance()
        (self.service_orchestrator.db_handler.
            update_node_instance_network_function_map) = mock.MagicMock(
                return_value=None)
        network_function = self.nfp_db.get_network_function(
            self.session, nfi['network_function_id'])
        self.assertEqual([nfi['id']],
                         network_function['network_function_instances'])
        data = {'network_function_instance': nfi}
        test_event = Event(data=data)
        test_event.context = data
        self.service_orchestrator.delete_network_function_instance(
            test_event)
        db_nfi = self.nfp_db.get_network_function_instance(
            self.session, nfi['id'])
        self.assertEqual(nfp_constants.PENDING_DELETE, db_nfi['status'])

    def test_event_handle_device_deleted(self):
        nfi = self.create_network_function_instance()
        ns_id = nfi['network_function_id']
        request_data = {'network_function_instance_id': nfi['id'],
                        'service_type': 'service_type'}
        test_event = Event(data=request_data)
        self.service_orchestrator.handle_device_deleted(
            test_event)
        self.assertRaises(nfp_exc.NetworkFunctionInstanceNotFound,
                          self.nfp_db.get_network_function_instance,
                          self.session,
                          nfi['id'])
        self.assertRaises(nfp_exc.NetworkFunctionNotFound,
                          self.nfp_db.get_network_function,
                          self.session,
                          ns_id)

    @mock.patch.object(
        nso.ServiceOrchestrator, "_create_event")
    @mock.patch.object(
        openstack_driver.KeystoneClient, "get_admin_token")
    @mock.patch.object(
        openstack_driver.GBPClient, "get_service_profile")
    @mock.patch.object(
        nso.NSOConfiguratorRpcApi, "policy_target_add_user_config")
    def test_handle_policy_target_added(self, mock_create_rpc,
                                        mock_get_service_profile,
                                        mock_get_admin_token,
                                        mock_create_event):
        (self.service_orchestrator.db_handler.
            update_node_instance_network_function_map) = mock.MagicMock(
            return_value=None)
        nfi = self.create_network_function_instance()
        network_function_id = nfi['network_function_id']
        policy_target = mock.Mock()
        transport.parse_service_flavor_string = mock.MagicMock(
            return_value={'device_type': 'VM',
                          'service_vendor': 'vyos'})
        with mock.patch.object(
                self.service_orchestrator.config_driver,
                "handle_policy_target_operations") as\
                mock_handle_policy_target_operations:
            mock_handle_policy_target_operations.return_value = 'stack_id'
            self.service_orchestrator.handle_policy_target_added(
                self.context, network_function_id, policy_target)
        db_nf = self.nfp_db.get_network_function(
            self.session, nfi['network_function_id'])
        self.assertIsNotNone(db_nf['config_policy_id'])

    @mock.patch.object(
        nso.ServiceOrchestrator, "_create_event")
    @mock.patch.object(
        openstack_driver.KeystoneClient, "get_admin_token")
    @mock.patch.object(
        openstack_driver.GBPClient, "get_service_profile")
    @mock.patch.object(
        nso.NSOConfiguratorRpcApi, "policy_target_remove_user_config")
    def test_handle_policy_target_removed(self, mock_delete_rpc,
                                          mock_get_service_profile,
                                          mock_get_admin_token,
                                          mock_create_event):
        (self.service_orchestrator.db_handler.
            update_node_instance_network_function_map) = mock.MagicMock(
            return_value=None)
        nfi = self.create_network_function_instance()
        network_function_id = nfi['network_function_id']
        policy_target = mock.Mock()
        transport.parse_service_flavor_string = mock.MagicMock(
            return_value={'device_type': 'VM',
                          'service_vendor': 'vyos'})
        with mock.patch.object(
                self.service_orchestrator.config_driver,
                "handle_policy_target_operations") as mock_handle_pt_removed:
            mock_handle_pt_removed.return_value = 'stack_id'
            self.service_orchestrator.handle_policy_target_removed(
                self.context, network_function_id, policy_target)
        db_nf = self.nfp_db.get_network_function(
            self.session, nfi['network_function_id'])
        self.assertIsNotNone(db_nf['config_policy_id'])

    @mock.patch.object(
        nso.ServiceOrchestrator, "_create_event")
    @mock.patch.object(
        openstack_driver.KeystoneClient, "get_admin_token")
    @mock.patch.object(
        openstack_driver.GBPClient, "get_service_profile")
    @mock.patch.object(
        nso.NSOConfiguratorRpcApi, "consumer_add_user_config")
    def test_handle_consumer_ptg_added(self, mock_create_rpc,
                                       mock_get_service_profile,
                                       mock_get_admin_token,
                                       mock_create_event):
        (self.service_orchestrator.db_handler.
            update_node_instance_network_function_map) = mock.MagicMock(
                return_value=None)
        nfi = self.create_network_function_instance()
        network_function_id = nfi['network_function_id']
        policy_target_group = mock.Mock()
        transport.parse_service_flavor_string = mock.MagicMock(
            return_value={'device_type': 'VM',
                          'service_vendor': 'vyos'})
        with mock.patch.object(
                self.service_orchestrator.config_driver,
                "handle_consumer_ptg_operations") as\
                mock_handle_consumer_ptg_added:
            mock_handle_consumer_ptg_added.return_value = 'stack_id'
            nfp_context.init()

            self.service_orchestrator.handle_consumer_ptg_added(
                self.context, network_function_id, policy_target_group)
        db_nf = self.nfp_db.get_network_function(
            self.session, nfi['network_function_id'])
        tag_str = 'heat_config'
        self.assertIsNotNone(db_nf['config_policy_id'])
        service_config = db_nf['service_config']
        network_function_details = (
            self.service_orchestrator.get_network_function_details(
                db_nf['id']))
        network_function_details['network_function'][
            'status'] = 'status'
        network_function_data = {
            'service_type': mock.ANY,
            'network_function_details': network_function_details,
            'consumer_ptg': policy_target_group
        }
        mock_create_rpc.assert_called_once_with(
            network_function_data, service_config, tag_str
        )

    @mock.patch.object(
        nso.ServiceOrchestrator, "_create_event")
    @mock.patch.object(
        openstack_driver.KeystoneClient, "get_admin_token")
    @mock.patch.object(
        openstack_driver.GBPClient, "get_service_profile")
    @mock.patch.object(
        nso.NSOConfiguratorRpcApi, "consumer_remove_user_config")
    def test_handle_consumer_ptg_removed(self, mock_delete_rpc,
                                         mock_get_service_profile,
                                         mock_get_admin_token,
                                         mock_create_event):
        (self.service_orchestrator.db_handler.
            update_node_instance_network_function_map) = mock.MagicMock(
                return_value=None)
        nfi = self.create_network_function_instance()
        network_function_id = nfi['network_function_id']
        policy_target_group = mock.Mock()
        transport.parse_service_flavor_string = mock.MagicMock(
            return_value={'device_type': 'VM',
                          'service_vendor': 'vyos'})
        with mock.patch.object(
                self.service_orchestrator.config_driver,
                "handle_consumer_ptg_operations") as\
                mock_handle_consumer_ptg_removed:
            mock_handle_consumer_ptg_removed.return_value = 'stack_id'
            nfp_context.init()
            self.service_orchestrator.handle_consumer_ptg_removed(
                self.context, network_function_id, policy_target_group)
        db_nf = self.nfp_db.get_network_function(
            self.session, nfi['network_function_id'])
        tag_str = 'heat_config'
        self.assertIsNotNone(db_nf['config_policy_id'])
        service_config = db_nf['service_config']
        network_function_details = (
            self.service_orchestrator.get_network_function_details(
                db_nf['id']))
        network_function_details['network_function'][
            'status'] = 'status'
        network_function_data = {
            'service_type': mock.ANY,
            'network_function_details': network_function_details,
            'consumer_ptg': policy_target_group
        }
        mock_delete_rpc.assert_called_once_with(
            network_function_data, service_config, tag_str
        )



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.db import model_base
from oslo_log import log
import sqlalchemy as sa
from sqlalchemy.ext import baked

LOG = log.getLogger(__name__)

BAKERY = baked.bakery(_size_alert=lambda c: LOG.warning(
    "sqlalchemy baked query cache size exceeded in %s", __name__))


class ApicSegmentationLabelDB(model_base.BASEV2):
    __tablename__ = 'gp_apic_mapping_segmentation_labels'
    policy_target_id = sa.Column(
        sa.String(36), sa.ForeignKey('gp_policy_targets.id',
                                     ondelete="CASCADE"), primary_key=True)
    segmentation_label = sa.Column(sa.String(255), primary_key=True)


class ApicSegmentationLabelDBMixin(object):

    def get_policy_target_segmentation_labels(self, session, policy_target_id):
        query = BAKERY(lambda s: s.query(
            ApicSegmentationLabelDB))
        query += lambda q: q.filter_by(
            policy_target_id=sa.bindparam('policy_target_id'))
        rows = query(session).params(
            policy_target_id=policy_target_id).all()

        return rows

    def get_policy_target_segmentation_label(self, session, policy_target_id,
                                             segmentation_label):
        query = BAKERY(lambda s: s.query(
            ApicSegmentationLabelDB))
        query += lambda q: q.filter_by(
            policy_target_id=sa.bindparam('policy_target_id'),
            segmentation_label=sa.bindparam('segmentation_label'))
        row = query(session).params(
            policy_target_id=policy_target_id,
            segmentation_label=segmentation_label).one()

        return row

    def add_policy_target_segmentation_label(self, session, policy_target_id,
                                             segmentation_label):
        row = ApicSegmentationLabelDB(policy_target_id=policy_target_id,
                                      segmentation_label=segmentation_label)
        session.add(row)

    def delete_policy_target_segmentation_label(
        self, session, policy_target_id, segmentation_label):
        row = self.get_policy_target_segmentation_label(
            session, policy_target_id, segmentation_label)
        session.delete(row)



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

import collections
import subprocess

from oslo_log import log as logging
from tempest.common.utils import data_utils
from tempest import config
from tempest.scenario import manager
from tempest import test

CONF = config.CONF
LOG = logging.getLogger(__name__)
Floating_IP_tuple = collections.namedtuple('Floating_IP_tuple',
                                           ['floating_ip', 'server'])
MAX_RETRY = 50


class TestServerEp(manager.NetworkScenarioTest):
    def setUp(self):
        super(TestServerEp, self).setUp()
        self.keypairs = {}
        self.servers = []
        self.deleted_rule = False
        self.track_success = False

    def ping_remote(self, should_connect, should_check_floating_ip_status):
        floating_ip, server = self.floating_ip_tuple
        ip_address = floating_ip['floating_ip_address']
        private_key = None
        floatingip_status = 'DOWN'

        if should_connect:
            private_key = self._get_server_key(server)
            if private_key is not None:
                floatingip_status = 'ACTIVE'

        # Check FloatingIP Status before initiating a connection
        if should_check_floating_ip_status:
            self.check_floating_ip_status(floating_ip, floatingip_status)
        LOG.debug('Starting Ping to Connect')
        self._track_connectivity(ip_address)
        return self.track_success

    def _track_connectivity(self, floating_ip):
        ping_command = "ping -O " + floating_ip
        process = subprocess.Popen(
            ping_command, shell=True, stdout=subprocess.PIPE, close_fds=True)
        LOG.debug('Tracking Connectivity')
        error_msg = "no answer yet for icmp_seq"
        success_msg = "bytes from " + floating_ip
        retry = MAX_RETRY

        # To skip first line of ping command
        output = process.stdout.readline()

        while process.poll() is None and retry > 0:
            output = process.stdout.readline()
            output = output.strip()
            if not output:
                LOG.error('Ping failed - Nothing to show')
                break
            if error_msg in output and self.deleted_rule is True:
                self.track_success = True
                LOG.debug("Rules Deleted Successfully")
                break
            elif success_msg in output and self.deleted_rule is True:
                LOG.debug("Waiting for rule to be deleted in fabric")
                retry -= 1
                if retry is 0:
                    LOG.error("Error - Still Pinging even after "
                              "deleting the security group rule")
                    break
            elif success_msg in output and self.deleted_rule is False:
                LOG.debug("Waiting for rules to be "
                          "deleted from security group")
                self.deleted_rule = self._delete_security_group_rule(
                                    self.servers)
                retry -= 1
                if retry is 0:
                    LOG.error("Security rule was not deleted in time")
                    break
            else:
                LOG.debug("Waiting for Server to get Active")
                retry -= 1
                if retry is 0:
                    LOG.error("Error - %s", (output,))
                    break

        LOG.debug("Closing Subprocess")
        process.stdout.close()
        process.kill()

    def _delete_security_group_rule(self, servers):
        for server in servers:
            sg = server['security_groups']
            LOG.debug("Security group is %s ", (sg,))
            rule_list_body = (
                self.security_group_rules_client.list_security_group_rules())
            for rule in rule_list_body['security_group_rules']:
                if ("icmp" == rule["protocol"] and
                        rule["direction"] == "ingress"):
                    LOG.debug("The rule is %s ", (rule,))
                    (self.security_group_rules_client.
                     delete_security_group_rule(rule["id"]))

        return True

    def _setup_network_and_servers(self, **kwargs):
        boot_with_port = kwargs.pop('boot_with_port', False)
        self.security_group = self._create_security_group(
                              tenant_id=self.tenant_id)
        security_groups = [{'name': self.security_group['name']}]
        self.network, self.subnet, self.router = self.create_networks(**kwargs)
        self.check_networks()

        self.ports = []
        self.port_id = None
        if boot_with_port:
            # create a port on the network and boot with that
            self.port_id = self._create_port(self.network['id'])['id']
            self.ports.append({'port': self.port_id})

        name = data_utils.rand_name('server-smoke')
        server = self._create_server(name, self.network,
                 security_groups, self.port_id)

        floating_ip = self.create_floating_ip(server)
        self.floating_ip_tuple = Floating_IP_tuple(floating_ip, server)

    def check_networks(self):
        """Checks that we see the newly created network/subnet/router

        via checking the result of list_[networks,routers,subnets]
        """

        seen_nets = self._list_networks()
        seen_names = [n['name'] for n in seen_nets]
        seen_ids = [n['id'] for n in seen_nets]
        self.assertIn(self.network['name'], seen_names)
        self.assertIn(self.network['id'], seen_ids)

        if self.subnet:
            seen_subnets = self._list_subnets()
            seen_net_ids = [n['network_id'] for n in seen_subnets]
            seen_subnet_ids = [n['id'] for n in seen_subnets]
            self.assertIn(self.network['id'], seen_net_ids)
            self.assertIn(self.subnet['id'], seen_subnet_ids)

        if self.router:
            seen_routers = self._list_routers()
            seen_router_ids = [n['id'] for n in seen_routers]
            seen_router_names = [n['name'] for n in seen_routers]
            self.assertIn(self.router['name'],
                          seen_router_names)
            self.assertIn(self.router['id'],
                          seen_router_ids)

    def _create_server(self, name, network, security_groups, port_id=None):
        keypair = self.create_keypair()
        self.keypairs[keypair['name']] = keypair
        network = {'uuid': network['id']}
        if port_id is not None:
            network['port'] = port_id

        server = self.create_server(
            name=name,
            networks=[network],
            key_name=keypair['name'],
            security_groups=security_groups,
            wait_until='ACTIVE')
        self.servers.append(server)
        return server

    def _get_server_key(self, server):
        return self.keypairs[server['key_name']]['private_key']

    @test.attr(type='smoke')
    @test.services('compute', 'network')
    def test_server_ep(self):
        self._setup_network_and_servers()
        self.assertTrue(self.ping_remote(True, True))
        LOG.debug("Testing Done")



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


class NFPNetworkDriverBase(object):
    """ NFP Network Driver Base class

    Handles ports, operations on them
    """

    def __init__(self):
        pass

    def setup_traffic_steering(self):
        pass

    def create_port(self, token, admin_id, net_id, name=None):
        pass

    def delete_port(self, token, port_id):
        pass

    def get_port_id(self, token, port_id):
        pass

    def get_port_details(self, token, port_id):
        pass

    def set_promiscuos_mode(self, token, port_id, enable_port_security):
        pass
